"""
Transcript tile processing module for spatial transcriptomics data.
"""

import concurrent.futures
from pathlib import Path

import numpy as np
import polars as pl
from scipy.sparse import csr_matrix
from tqdm import tqdm

from .boundary_tile import _get_name_mapping


def _process_coarse_tile_transcripts(
    trx,
    i,
    j,
    coarse_tile_x_min,
    coarse_tile_x_max,
    coarse_tile_y_min,
    coarse_tile_y_max,
    tile_size,
    path_trx_tiles,
    x_min,
    y_min,
    n_fine_tiles_x,
    n_fine_tiles_y,
    max_workers=1,
):
    """
    Process a single coarse tile for transcript data.

    Parameters:
    - trx: Transcript data
    - i, j: Coarse tile indices
    - coarse_tile_x_min, coarse_tile_x_max: X bounds for coarse tile
    - coarse_tile_y_min, coarse_tile_y_max: Y bounds for coarse tile
    - tile_size: Size of each tile
    - path_trx_tiles: Path to transcript tiles
    - x_min, y_min: Minimum coordinates
    - n_fine_tiles_x, n_fine_tiles_y: Number of fine tiles
    - max_workers: Maximum number of workers
    """
    # Filter the entire dataset for the current coarse tile
    coarse_tile = trx.filter(
        (pl.col("transformed_x") >= coarse_tile_x_min)
        & (pl.col("transformed_x") < coarse_tile_x_max)
        & (pl.col("transformed_y") >= coarse_tile_y_min)
        & (pl.col("transformed_y") < coarse_tile_y_max)
    )

    if not coarse_tile.is_empty():
        # Now process fine tiles using global fine tile indices
        _process_fine_tiles_transcripts(
            coarse_tile,
            i,
            j,
            coarse_tile_x_min,
            coarse_tile_x_max,
            coarse_tile_y_min,
            coarse_tile_y_max,
            tile_size,
            path_trx_tiles,
            x_min,
            y_min,
            n_fine_tiles_x,
            n_fine_tiles_y,
            max_workers,
        )


def _process_fine_tiles_transcripts(
    coarse_tile,
    coarse_i,
    coarse_j,
    coarse_tile_x_min,
    coarse_tile_x_max,
    coarse_tile_y_min,
    coarse_tile_y_max,
    tile_size,
    path_trx_tiles,
    x_min,
    y_min,
    n_fine_tiles_x,
    n_fine_tiles_y,
    max_workers=1,
):
    """
    Process fine tiles within a coarse tile for transcript data.

    Parameters:
    - coarse_tile: Coarse tile data
    - coarse_i, coarse_j: Coarse tile indices
    - coarse_tile_x_min, coarse_tile_x_max: X bounds for coarse tile
    - coarse_tile_y_min, coarse_tile_y_max: Y bounds for coarse tile
    - tile_size: Size of each tile
    - path_trx_tiles: Path to transcript tiles
    - x_min, y_min: Minimum coordinates
    - n_fine_tiles_x, n_fine_tiles_y: Number of fine tiles
    - max_workers: Maximum number of workers
    """
    # Use ThreadPoolExecutor for parallel processing of fine-grain tiles within the coarse tile
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []

        # Iterate over fine-grain tiles within the global bounds
        for fine_i in range(n_fine_tiles_x):
            fine_tile_x_min = x_min + fine_i * tile_size
            fine_tile_x_max = fine_tile_x_min + tile_size

            # Process only if the fine tile falls within the current coarse tile's bounds
            if fine_tile_x_min < coarse_tile_x_min or fine_tile_x_max > coarse_tile_x_max:
                continue

            for fine_j in range(n_fine_tiles_y):
                fine_tile_y_min = y_min + fine_j * tile_size
                fine_tile_y_max = fine_tile_y_min + tile_size

                # Process only if the fine tile falls within the current coarse tile's bounds
                if not (
                    fine_tile_y_min >= coarse_tile_y_min and fine_tile_y_max <= coarse_tile_y_max
                ):
                    continue

                # Submit the task for each fine tile to process in parallel
                futures.append(
                    executor.submit(
                        _filter_and_save_fine_tile,
                        coarse_tile,
                        coarse_i,
                        coarse_j,
                        fine_i,
                        fine_j,
                        fine_tile_x_min,
                        fine_tile_x_max,
                        fine_tile_y_min,
                        fine_tile_y_max,
                        path_trx_tiles,
                    )
                )

        # Wait for all futures to complete
        for future in concurrent.futures.as_completed(futures):
            future.result()  # Raise exceptions if any occurred during execution


def _filter_and_save_fine_tile(
    coarse_tile,
    coarse_i,
    coarse_j,
    fine_i,
    fine_j,
    fine_tile_x_min,
    fine_tile_x_max,
    fine_tile_y_min,
    fine_tile_y_max,
    path_trx_tiles,
):
    """
    Filter and save a fine tile for transcript data.

    Parameters:
    - coarse_tile: Coarse tile data
    - coarse_i, coarse_j: Coarse tile indices
    - fine_i, fine_j: Fine tile indices
    - fine_tile_x_min, fine_tile_x_max: X bounds for fine tile
    - fine_tile_y_min, fine_tile_y_max: Y bounds for fine tile
    - path_trx_tiles: Path to transcript tiles
    """
    # Filter the coarse tile for the current fine tile's boundaries
    fine_tile_trx = coarse_tile.filter(
        (pl.col("transformed_x") >= fine_tile_x_min)
        & (pl.col("transformed_x") < fine_tile_x_max)
        & (pl.col("transformed_y") >= fine_tile_y_min)
        & (pl.col("transformed_y") < fine_tile_y_max)
    )

    if not fine_tile_trx.is_empty():
        # Add geometry column as a list of [x, y] pairs
        fine_tile_trx = fine_tile_trx.with_columns(
            pl.concat_list([pl.col("transformed_x"), pl.col("transformed_y")]).alias("geometry")
        ).drop(["transformed_x", "transformed_y", "cell_id", "transcript_id"])

        # Define the filename based on fine tile coordinates
        filename = Path(path_trx_tiles) / f"transcripts_tile_{fine_i}_{fine_j}.parquet"

        # Save the filtered DataFrame to a Parquet file
        fine_tile_trx.to_pandas().to_parquet(filename, index=False)


def _load_transcript_data_by_technology(technology, path_trx):
    """
    Load transcript data based on technology.

    Parameters:
    - technology: Technology type
    - path_trx: Path to transcript data

    Returns:
    - Polars DataFrame with transcript data
    """
    if technology == "MERSCOPE":
        trx_ini = pl.read_csv(path_trx, columns=["gene", "global_x", "global_y"])
        return trx_ini.with_columns(
            [
                pl.col("cell_id"),
                pl.col("transcript_id"),
                pl.col("global_x").alias("x"),
                pl.col("global_y").alias("y"),
                pl.col("gene").alias("name"),
            ]
        ).select(["name", "x", "y"])

    if technology == "Xenium":
        return pl.read_parquet(path_trx).select(
            [
                pl.col("cell_id"),
                pl.col("transcript_id"),
                pl.col("feature_name").alias("name"),
                pl.col("x_location").alias("x"),
                pl.col("y_location").alias("y"),
            ]
        )

    raise ValueError(f"Unsupported technology: {technology}")


def _apply_gene_mapping(trx_ini, gene_str_to_int_mapping):
    """
    Apply gene name mapping to transcript data.

    Parameters:
    - trx_ini: Initial transcript data
    - gene_str_to_int_mapping: Mapping dictionary for gene names

    Returns:
    - Transcript data with mapped gene names
    """
    # Create a list with the mapped names; if a name isn't in gene_map, keep the original.
    mapped_names = [gene_str_to_int_mapping.get(name, name) for name in trx_ini["name"]]

    # Replace the "name" column using with_columns.
    return trx_ini.with_columns([pl.Series("name", mapped_names)])


def _transform_coordinates_in_chunks(trx_ini, chunk_size, transformation_matrix, image_scale):
    """
    Transform coordinates in chunks for memory efficiency.

    Parameters:
    - trx_ini: Initial transcript data
    - chunk_size: Size of chunks to process
    - transformation_matrix: Transformation matrix
    - image_scale: Image scale factor

    Returns:
    - Concatenated transformed data
    """
    all_chunks = []
    sparse_matrix = csr_matrix(transformation_matrix)

    for start_row in tqdm(range(0, trx_ini.height, chunk_size), desc="Processing chunks"):
        chunk = trx_ini.slice(start_row, chunk_size)

        points = np.hstack([chunk.select(["x", "y"]).to_numpy(), np.ones((chunk.height, 1))])
        transformed_points = sparse_matrix.dot(points.T).T[:, :2]

        # Create new transformed columns and drop original x, y columns
        transformed_chunk = chunk.with_columns(
            [
                (pl.Series(transformed_points[:, 0]) * image_scale).round(2).alias("transformed_x"),
                (pl.Series(transformed_points[:, 1]) * image_scale).round(2).alias("transformed_y"),
            ]
        ).drop(["x", "y"])
        all_chunks.append(transformed_chunk)

    # Concatenate all chunks after processing
    return pl.concat(all_chunks)


def transform_transcript_coordinates(
    technology,
    path_trx,
    chunk_size,
    transformation_matrix,
    image_scale=1,
    gene_str_to_int_mapping=None,
):
    """
    Transform transcript coordinates from microns to image space.

    Parameters:
    - technology: Technology type
    - path_trx: Path to transcript data
    - chunk_size: Chunk size for processing
    - transformation_matrix: Transformation matrix
    - image_scale: Image scale factor
    - gene_str_to_int_mapping: Gene name mapping dictionary

    Returns:
    - Transformed transcript data
    """
    if gene_str_to_int_mapping is None:
        gene_str_to_int_mapping = {}

    # Load the transcript data based on the technology using Polars
    trx_ini = _load_transcript_data_by_technology(technology, path_trx)

    # Apply gene mapping
    trx_ini = _apply_gene_mapping(trx_ini, gene_str_to_int_mapping)

    # Process the data in chunks and apply transformations
    return _transform_coordinates_in_chunks(trx_ini, chunk_size, transformation_matrix, image_scale)


def _get_custom_tile_bounds(path_trx):
    """
    Get tile bounds for custom technology.

    Parameters:
    - path_trx: Path to transcript data

    Returns:
    - Tuple of (x_min, y_min, x_max, y_max)
    """
    x_min, y_min = 0, 0
    x_max, y_max = (
        pl.read_parquet(path_trx)
        .select(
            [
                pl.col("x_image_coords").max().alias("x_max"),
                pl.col("y_image_coords").max().alias("y_max"),
            ]
        )
        .row(0)
    )
    return x_min, y_min, x_max, y_max


def _get_transformed_tile_bounds(trx):
    """
    Get tile bounds for transformed transcript data.

    Parameters:
    - trx: Transformed transcript data

    Returns:
    - Tuple of (x_min, y_min, x_max, y_max)
    """
    x_min, y_min = 0, 0
    x_max, y_max = trx.select(
        [
            pl.col("transformed_x").max().alias("x_max"),
            pl.col("transformed_y").max().alias("y_max"),
        ]
    ).row(0)
    return x_min, y_min, x_max, y_max


def _process_transcript_tiles_parallel(
    trx, x_min, y_min, x_max, y_max, tile_size, coarse_tile_factor, path_trx_tiles, max_workers
):
    """
    Process transcript tiles in parallel.

    Parameters:
    - trx: Transcript data
    - x_min, y_min, x_max, y_max: Coordinate bounds
    - tile_size: Size of each tile
    - coarse_tile_factor: Coarse tile scaling factor
    - path_trx_tiles: Path to transcript tiles
    - max_workers: Maximum number of workers
    """
    # Calculate the number of fine-grain tiles globally
    n_fine_tiles_x = int(np.ceil((x_max - x_min) / tile_size))
    n_fine_tiles_y = int(np.ceil((y_max - y_min) / tile_size))

    # Calculate the number of coarse-grain tiles
    n_coarse_tiles_x = int(np.ceil((x_max - x_min) / (coarse_tile_factor * tile_size)))
    n_coarse_tiles_y = int(np.ceil((y_max - y_min) / (coarse_tile_factor * tile_size)))

    # Use ThreadPoolExecutor for parallel processing of coarse-grain tiles
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for i in range(n_coarse_tiles_x):
            coarse_tile_x_min = x_min + i * (coarse_tile_factor * tile_size)
            coarse_tile_x_max = coarse_tile_x_min + (coarse_tile_factor * tile_size)

            for j in range(n_coarse_tiles_y):
                coarse_tile_y_min = y_min + j * (coarse_tile_factor * tile_size)
                coarse_tile_y_max = coarse_tile_y_min + (coarse_tile_factor * tile_size)

                # Submit each coarse tile for parallel processing
                futures.append(
                    executor.submit(
                        _process_coarse_tile_transcripts,
                        trx,
                        i,
                        j,
                        coarse_tile_x_min,
                        coarse_tile_x_max,
                        coarse_tile_y_min,
                        coarse_tile_y_max,
                        tile_size,
                        path_trx_tiles,
                        x_min,
                        y_min,
                        n_fine_tiles_x,
                        n_fine_tiles_y,
                        max_workers,
                    )
                )

        # Wait for all coarse tiles to complete
        for future in tqdm(
            concurrent.futures.as_completed(futures),
            desc="Processing coarse tiles",
            unit="tile",
        ):
            future.result()  # Raise exceptions if any occurred during execution


def make_trx_tiles(
    technology,
    path_trx,
    path_transformation_matrix=None,
    path_trx_tiles=None,
    coarse_tile_factor=10,
    tile_size=250,
    chunk_size=1000000,
    verbose=False,
    image_scale=1,
    max_workers=1,
):
    """
    Processes transcript data by dividing it into coarse-grain and fine-grain tiles,
    applying transformations, and saving the results in a parallelized manner.

    Parameters
    ----------
    technology : str
        The technology used for generating the transcript data (e.g., "MERSCOPE" or "Xenium").
    path_trx : str
        Path to the file containing the transcript data.
    path_transformation_matrix : str
        Path to the file containing the transformation matrix (CSV file).
    path_trx_tiles : str
        Directory path where the output files (Parquet files) for each tile will be saved.
    coarse_tile_factor : int, optional
        Scaling factor of each coarse-grain tile comparing to the fine tile size.
    tile_size : int, optional
        Size of each fine-grain tile in microns (default is 250).
    chunk_size : int, optional
        Number of rows to process per chunk for memory efficiency (default is 1000000).
    verbose : bool, optional
        Flag to enable verbose output (default is False).
    image_scale : float, optional
        Scale factor to apply to the transcript coordinates (default is 0.5).
    max_workers : int, optional
        Maximum number of parallel workers for processing tiles (default is 1).

    Returns
    -------
    dict
        A dictionary containing the bounds of the processed data in both x and y directions.
    """
    if technology == "custom":
        x_min, y_min, x_max, y_max = _get_custom_tile_bounds(path_trx)
    else:
        # Create output directory
        tiles_path = Path(path_trx_tiles)
        tiles_path.mkdir(exist_ok=True)

        transformation_matrix = np.loadtxt(path_transformation_matrix)

        gene_str_to_int_mapping = _get_name_mapping(
            str(Path(path_transformation_matrix).parent),
            layer="transcript",
        )

        trx = transform_transcript_coordinates(
            technology,
            path_trx,
            chunk_size,
            transformation_matrix,
            image_scale,
            gene_str_to_int_mapping=gene_str_to_int_mapping,
        )

        # Get min and max x, y values
        x_min, y_min, x_max, y_max = _get_transformed_tile_bounds(trx)

        # Process tiles in parallel
        _process_transcript_tiles_parallel(
            trx,
            x_min,
            y_min,
            x_max,
            y_max,
            tile_size,
            coarse_tile_factor,
            path_trx_tiles,
            max_workers,
        )

    # Return the tile bounds
    return {
        "x_min": x_min,
        "x_max": x_max,
        "y_min": y_min,
        "y_max": y_max,
    }


# Legacy function names for backward compatibility
process_coarse_tile = _process_coarse_tile_transcripts
process_fine_tiles = _process_fine_tiles_transcripts
filter_and_save_fine_tile = _filter_and_save_fine_tile



def cluster_row_and_col(
    net,
    dist_type="cosine",
    linkage_type="average",
    dendro=True,
    run_clustering=True,
    run_rank=True,
    ignore_cat=False,
    calc_cat_pval=False,
    links=False,
    clust_library="scipy",
    min_samples=1,
    min_cluster_size=2,
):
    """cluster net.dat and make visualization json, net.viz.
    optionally leave out dendrogram colorbar groups with dendro argument"""

    # import umap
    from copy import deepcopy

    from . import cat_pval, categories, make_viz

    dm = {}
    for axis in ["row", "col"]:
        # save directly to dat structure
        node_info = net.dat["node_info"][axis]

        node_info["ini"] = list(range(len(net.dat["nodes"][axis]), -1, -1))

        tmp_mat = deepcopy(net.dat["mat"])

        # calc distance matrix
        if clust_library != "hdbscan":
            dm[axis] = calc_distance_matrix(tmp_mat, axis, dist_type)
        else:
            dm[axis] = None

        # dm[axis] = calc_distance_matrix(tmp_mat, axis, dist_type)

        # cluster
        if run_clustering:
            node_info["clust"], node_info["Y"] = clust_and_group(
                net,
                dm[axis],
                axis,
                tmp_mat,
                dist_type=dist_type,
                linkage_type=linkage_type,
                clust_library=clust_library,
                min_samples=min_samples,
                min_cluster_size=min_cluster_size,
            )
        else:
            dendro = False
            node_info["clust"] = node_info["ini"]

        # sorting
        if run_rank:
            node_info["rank"] = sort_rank_nodes(net, axis, "sum")
            node_info["rankvar"] = sort_rank_nodes(net, axis, "var")
        else:
            node_info["rank"] = node_info["ini"]
            node_info["rankvar"] = node_info["ini"]

        ##################################
        if not ignore_cat:
            categories.calc_cat_clust_order(net, axis)

    if calc_cat_pval:
        cat_pval.main(net)

    # make the visualization json
    make_viz.viz_json(net, dendro, links)

    return dm


def calc_distance_matrix(tmp_mat, axis, dist_type="cosine"):
    from scipy.spatial.distance import pdist

    if axis == "row":
        inst_dm = pdist(tmp_mat, metric=dist_type)
    elif axis == "col":
        inst_dm = pdist(tmp_mat.transpose(), metric=dist_type)

    inst_dm[inst_dm < 0] = float(0)

    return inst_dm


def clust_and_group(
    net,
    inst_dm,
    axis,
    mat,
    dist_type="cosine",
    linkage_type="average",
    clust_library="scipy",
    min_samples=1,
    min_cluster_size=2,
):
    # print(clust_library)

    import pandas as pd
    import scipy.cluster.hierarchy as hier

    if clust_library == "scipy":
        Y = hier.linkage(inst_dm, method=linkage_type)

    elif clust_library == "fastcluster":
        import fastcluster

        Y = fastcluster.linkage(inst_dm, method=linkage_type)

    elif clust_library == "hdbscan":
        # print('HDBSCAN!')
        import hdbscan

        # pca-umap-hdbscan using data (no pre-cal distance matrix)
        ######################################################
        from sklearn.decomposition import PCA

        clusterer = hdbscan.HDBSCAN(min_samples=min_samples, min_cluster_size=min_cluster_size)

        # rows are the data points, cols are dimensions
        n_components = 50
        if axis == "row":
            low_d_mat = (
                PCA(n_components=n_components).fit_transform(mat)
                if mat.shape[1] > n_components
                else mat
            )

        elif axis == "col":
            low_d_mat = (
                PCA(n_components=n_components).fit_transform(mat.transpose())
                if mat.shape[0] > n_components
                else mat.transpose()
            )

        # run UMAP on low_d_mat (after PCA)
        # print('running umap!!!!!!!!!!!!!!!!!!!!!!!!!!')
        import umap

        umap_mat = umap.UMAP(
            metric=dist_type,
            n_neighbors=5,
            min_dist=0.0,
            n_components=2,
            random_state=42,
        ).fit_transform(low_d_mat)

        umap_df = pd.DataFrame(
            umap_mat.transpose(), index=["x", "y"], columns=net.dat["nodes"][axis]
        )

        net.umap[axis] = umap_df
        clusterer.fit(umap_mat)

        Y = clusterer.single_linkage_tree_.to_numpy()

    Z = hier.dendrogram(Y, no_plot=True)
    inst_clust_order = Z["leaves"]

    return inst_clust_order, Y


def sort_rank_nodes(net, rowcol, rank_type):
    from copy import deepcopy
    from operator import itemgetter

    import numpy as np

    tmp_nodes = deepcopy(net.dat["nodes"][rowcol])
    inst_mat = deepcopy(net.dat["mat"])

    sum_term = []
    for i, node_name in enumerate(tmp_nodes):
        inst_dict = {"name": node_name}

        if rowcol == "row":
            inst_dict["rank"] = (
                np.sum(inst_mat[i, :]) if rank_type == "sum" else np.var(inst_mat[i, :])
            )
        else:
            inst_dict["rank"] = (
                np.sum(inst_mat[:, i]) if rank_type == "sum" else np.var(inst_mat[:, i])
            )

        sum_term.append(inst_dict)

    sum_term = sorted(sum_term, key=itemgetter("rank"), reverse=False)

    tmp_sort_nodes = [inst_dict["name"] for inst_dict in sum_term]

    return [tmp_sort_nodes.index(inst_node) for inst_node in tmp_nodes]



"""
Module for clustering high-dimensional data.
"""

# ruff: noqa: E722

# The Celldega Matrix Vizualization Method is being built using the approaches
# and code adaptations from the Clustergrammer-GL library, which is available at
# github.com/ismms-himc/clustergrammer2
# and being used under the license
#
# MIT License

# Copyright (c) 2021 Nicolas Fernandez

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import contextlib
from copy import deepcopy
from itertools import combinations
import json
from pathlib import Path
import random

import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.spatial.distance import pdist
from scipy.stats import mannwhitneyu, ttest_ind
from sklearn.metrics import auc, confusion_matrix, pairwise_distances, roc_curve
import statsmodels.stats.multitest as smm

from . import (
    categories,
    data_formats,
    downsample_fun,
    enrichr_functions as enr_fun,
    export_data,
    iframe_web_app,
    initialize_net,
    load_data,
    load_vect_post,
    make_clust_fun,
    normalize_fun,
    run_filter,
)


def hc(df, filter_n_top=None, norm_col="total", norm_row="zscore"):
    """
    This function performs hierarchical clustering on the rows and columns of a
    DataFrame and returns the visualization JSON for the Celldega-Matrix method.
    This function is developed using the approaches and code adaptations from the
    Clustergrammer2 project.

    Args:
      df (pd.DataFrame): A Pandas DataFrame.
      filter_n_top (int): The number of top dimensions to keep after filtering.
      norm_col (str): The type of normalization to apply to the columns.
      norm_row (str): The type of normalization to apply to the rows.

    Returns:
      dict: The visualization JSON for the Celldega-Matrix method


    Example:
    See [Landscape-Matrix_Xenium](../../../examples/brief_notebooks/Landscape-Matrix_Xenium) notebook.
    """

    net = Network()

    net.load_df(df)

    if filter_n_top is not None:
        net.filter_n_top(axis="row", n_top=5000)

    if norm_col == "total":
        net.normalize(axis="col", norm_type="umi")

    if norm_row == "zscore":
        net.normalize(axis="row", norm_type="zscore")

    net.cluster()

    return net.viz


class Network:
    """
    Clustergrammer.py takes a matrix as input (either from a file of a Pandas DataFrame), normalizes/filters, hierarchically clusters, and produces the :ref:`visualization_json` for :ref:`clustergrammer_js`.

    Networks have two states:

      1. the data state, where they are stored as a matrix and nodes
      2. the viz state where they are stored as viz.links, viz.row_nodes, and viz.col_nodes.

    The goal is to start in a data-state and produce a viz-state of
    the network that will be used as input to clustergram.js.
    """

    def __init__(self, widget=None):
        initialize_net.main(self, widget)

    def reset(self):
        """
        This re-initializes the Network object.
        """
        initialize_net.main(self)

    def load_file(self, filename):
        """
        Load TSV file.
        """
        load_data.load_file(self, filename)

    def load_file_as_string(self, file_string, filename=""):
        """
        Load file as a string.
        """
        load_data.load_file_as_string(self, file_string, filename=filename)

    def load_stdin(self):
        """
        Load stdin TSV-formatted string.
        """
        load_data.load_stdin(self)

    def load_tsv_to_net(self, file_buffer, filename=None):
        """
        This will load a TSV matrix file buffer; this is exposed so that it will
        be possible to load data without having to read from a file.
        """
        load_data.load_tsv_to_net(self, file_buffer, filename)

    def load_vect_post_to_net(self, vect_post):
        """
        Load data in the vector format JSON.
        """
        load_vect_post.main(self, vect_post)

    def load_data_file_to_net(self, filename):
        """
        Load Clustergrammer's dat format (saved as JSON).
        """
        inst_dat = self.load_json_to_dict(filename)
        load_data.load_data_to_net(self, inst_dat)

    def cluster(
        self,
        dist_type="cosine",
        run_clustering=True,
        dendro=True,
        views=None,
        linkage_type="average",
        sim_mat=False,
        filter_sim=0.0,
        calc_cat_pval=False,
        run_enrichr=None,
        enrichrgram=None,
        clust_library="scipy",
        min_samples=1,
        min_cluster_size=2,
    ):
        """
        The main function performs hierarchical clustering, optionally generates
        filtered views (e.g. row-filtered views), and generates the :
        ``visualization_json``.

        Used to set views equal to ['N_row_sum', 'N_row_var']
        """
        if views is None:
            views = []

        initialize_net.viz(self)

        make_clust_fun.make_clust(
            self,
            dist_type=dist_type,
            run_clustering=run_clustering,
            dendro=dendro,
            requested_views=views,
            linkage_type=linkage_type,
            sim_mat=sim_mat,
            filter_sim=filter_sim,
            calc_cat_pval=calc_cat_pval,
            run_enrichr=run_enrichr,
            enrichrgram=enrichrgram,
            clust_library=clust_library,
            min_samples=min_samples,
            min_cluster_size=min_cluster_size,
        )

    def swap_nan_for_zero(self):
        """
        Swaps all NaN (numpy NaN) instances for zero.
        """
        self.dat["mat"][np.isnan(self.dat["mat"])] = 0

    def load_df(
        self,
        df_ini,
        meta_col=None,
        meta_row=None,
        col_cats=None,
        row_cats=None,
        is_downsampled=False,
        meta_ds_row=None,
        meta_ds_col=None,
    ):
        """
        Load Pandas DataFrame.
        """
        self.reset()

        # load dataframe
        df = deepcopy(df_ini)

        if is_downsampled:
            if meta_ds_col is not None:
                self.meta_ds_col = meta_ds_col
            if meta_ds_row is not None:
                self.meta_ds_row = meta_ds_row

        # define downsampled status
        self.is_downsampled = is_downsampled
        # print('load_df: is_downsampled', is_downsampled)

        if not hasattr(self, "meta_col") and not hasattr(self, "meta_row"):
            self.meta_cat = False

        # load metadata
        if isinstance(meta_col, pd.DataFrame):
            self.meta_col = meta_col

            self.col_cats = meta_col.columns.tolist() if col_cats is None else col_cats

            self.meta_cat = True

        if isinstance(meta_row, pd.DataFrame):
            self.meta_row = meta_row

            self.row_cats = meta_row.columns.tolist() if row_cats is None else row_cats

            self.meta_cat = True

        data_formats.df_to_dat(self, df, define_cat_colors=True)

    def export_df(self):
        """
        Export Pandas DataFrame/
        """
        df = data_formats.dat_to_df(self)

        # drop tuple categories if downsampling
        if self.is_downsampled:
            df.columns = self.dat["nodes"]["col"]
            df.index = self.dat["nodes"]["row"]

        return df

    def df_to_dat(self, df, define_cat_colors=False):
        """
        Load Pandas DataFrame (will be deprecated).
        """
        data_formats.df_to_dat(self, df, define_cat_colors)

    def set_matrix_colors(self, pos="red", neg="blue"):
        self.viz["matrix_colors"] = {"pos": pos, "neg": neg}

    def set_global_cat_colors(self, df_meta):
        for inst_name in df_meta.index.tolist():
            inst_color = df_meta.loc[inst_name, "color"]

            self.viz["global_cat_colors"][inst_name] = inst_color

    def set_cat_color(self, axis, cat_index, cat_name, inst_color):
        if axis == 0:
            axis = "row"
        if axis == 1:
            axis = "col"

        with contextlib.suppress(Exception):
            # process cat_index
            cat_index = cat_index - 1
            cat_index = f"cat-{cat_index}"

            self.viz["cat_colors"][axis][cat_index][cat_name] = inst_color

    def dat_to_df(self):
        """
        Export Pandas DataFrams (will be deprecated).
        """
        return data_formats.dat_to_df(self)

    def export_net_json(self, net_type="viz", indent="no-indent"):
        """
        Export dat or viz JSON.
        """
        return export_data.export_net_json(self, net_type, indent)

    def export_viz_to_widget(self, which_viz="viz"):
        """
        Export viz JSON, for use with clustergrammer_widget. Formerly method was
        named widget.
        """

        return export_data.export_net_json(self, which_viz, "no-indent")

    def widget(
        self,
        which_viz="viz",
        link_net=None,
        link_net_js=None,
        clust_library="scipy",
        min_samples=1,
        min_cluster_size=2,
    ):
        """
        Generate a widget visualization using the widget. The export_viz_to_widget
        method passes the visualization JSON to the instantiated widget, which is
        returned and visualized on the front-end.
        """
        # run clustering if necessary
        if len(self.viz["row_nodes"]) == 0:
            self.cluster(
                clust_library=clust_library,
                min_samples=min_samples,
                min_cluster_size=min_cluster_size,
            )

            # add manual_category to viz json
            if "manual_category" in self.dat:
                self.viz["manual_category"] = self.dat["manual_category"]

            # add pre-z-score data to viz
            if "pre_zscore" in self.dat:
                self.viz["pre_zscore"] = self.dat["pre_zscore"]

        self.widget_instance = self.widget_class(network=self.export_viz_to_widget(which_viz))

        # initialize manual category
        if "manual_category" in self.dat:
            manual_cat = {"col": {"col_cat_colors": self.viz["cat_colors"]["col"]["cat-0"]}}

            man_cat_name = self.dat["manual_category"]["col"]
            manual_cat["col"][man_cat_name] = self.meta_col[man_cat_name].to_dict()

            self.widget_instance.manual_cat = json.dumps(manual_cat)

            self.widget_instance.observe(self.get_manual_category, names="manual_cat")

        # add link (python)
        if link_net is not None:
            inst_link = widgets.link(
                (self.widget_instance, "manual_cat"), (link_net.widget_instance, "manual_cat")
            )
            self.widget_instance.link = inst_link

        # add jslink (JavaScript)
        if link_net_js is not None:
            inst_link = widgets.jslink(
                (self.widget_instance, "manual_cat"), (link_net_js.widget_instance, "manual_cat")
            )
            self.widget_instance.link = inst_link

        return self.widget_instance

    def widget_df(self):
        """
        Export a DataFrame from the front-end visualization. For instance, a user
        can filter to show only a single cluster using the dendrogram and then
        get a dataframe of this cluster using the widget_df method.
        """

        if hasattr(self, "widget_instance"):
            if self.widget_instance.mat_string != "":
                tmp_net = deepcopy(Network())

                df_string = self.widget_instance.mat_string

                tmp_net.load_file_as_string(df_string)

                return tmp_net.export_df()

            return self.export_df()

        if hasattr(self, "widget_class"):
            print("Please make the widget before exporting the widget DataFrame.")
            print("Do this using the widget method: net.widget()")
            return None

        print("Can not make widget because Network has no attribute widget_class")
        print(
            "Please instantiate Network with clustergrammer_widget using: Network(clustergrammer_widget)"
        )
        return None

    def write_json_to_file(self, net_type, filename, indent="no-indent"):
        """
        Save dat or viz as a JSON to file.
        """
        export_data.write_json_to_file(self, net_type, filename, indent)

    def write_matrix_to_tsv(self, filename=None, df=None):
        """
        Export data-matrix to file.
        """
        return export_data.write_matrix_to_tsv(self, filename, df)

    def filter_sum(self, threshold, take_abs=True, axis=None, inst_rc=None):
        """
        Filter a network's rows or columns based on the sum across rows or columns.
        """

        if axis is None:
            axis = inst_rc
            print("warning inst_rc argument will be deprecated, please use axis")

            if inst_rc is None:
                print("please provide axis argument")
                return

        inst_df = self.dat_to_df()
        if axis == "row":
            inst_df = run_filter.df_filter_row_sum(inst_df, threshold, take_abs)
        elif axis == "col":
            inst_df = run_filter.df_filter_col_sum(inst_df, threshold, take_abs)
        self.df_to_dat(inst_df)

    def filter_n_top(self, n_top, rank_type="sum", inst_rc=None, axis=None):
        """
        Filter the matrix rows or columns based on sum/variance, and only keep the top
        N.
        """

        if axis is None:
            axis = inst_rc
            print("warning inst_rc argument will be deprecated, please use axis")

            if inst_rc is None:
                print("please provide axis argument")
                return

        inst_df = self.dat_to_df()

        inst_df = run_filter.filter_n_top(inst_rc, inst_df, n_top, rank_type)

        self.df_to_dat(inst_df)

    def filter_threshold(self, threshold, num_occur=1, inst_rc=None, axis=None):
        """
        Filter the matrix rows or columns based on num_occur values being above a
        threshold (in absolute value).
        """

        if axis is None:
            axis = inst_rc
            print("warning inst_rc argument will be deprecated, please use axis")

            if inst_rc is None:
                print("please provide axis argument")
                return

        inst_df = self.dat_to_df()

        inst_df = run_filter.filter_threshold(inst_df, axis, threshold, num_occur)

        self.df_to_dat(inst_df)

    def filter_cat(self, axis, cat_index, cat_name):
        """
        Filter the matrix based on their category. cat_index is the index of the category, the first category has index=1.
        """
        run_filter.filter_cat(self, axis, cat_index, cat_name)

    def filter_names(self, axis, names):
        """
        Filter the visualization using row/column names. The function takes, axis ('row'/'col') and names, a list of strings.
        """
        run_filter.filter_names(self, axis, names)

    def clip(self, lower=None, upper=None):
        """
        Trim values at input thresholds using pandas function
        """
        df = self.export_df()
        df = df.clip(lower=lower, upper=upper)
        self.load_df(df)

    def normalize(self, df=None, norm_type="zscore", axis="row", z_clip=None):
        """
        Normalize the matrix rows or columns using Z-score (zscore) or Quantile Normalization (qn). Users can optionally pass in a DataFrame to be normalized (and this will be incorporated into the Network object).
        """
        normalize_fun.run_norm(self, df, norm_type, axis, z_clip)

    def downsample(
        self,
        df=None,
        ds_type="kmeans",
        axis="row",
        num_samples=100,
        random_state=1000,
        ds_name="Downsample",
        ds_cluster_name="cluster",
    ):
        """
        Downsample the matrix rows or columns (currently supporting kmeans only). Users can optionally pass in a DataFrame to be downsampled (and this will be incorporated into the network object).
        """
        return downsample_fun.main(
            self, df, ds_type, axis, num_samples, random_state, ds_name, ds_cluster_name
        )

    def random_sample(
        self, num_samples, df=None, replace=False, weights=None, random_state=100, axis="row"
    ):
        """
        Return random sample of matrix.
        """

        if df is None:
            df = self.dat_to_df()

        if axis == "row":
            axis = 0
        if axis == "col":
            axis = 1

        df = self.export_df()
        df = df.sample(
            n=num_samples, replace=replace, weights=weights, random_state=random_state, axis=axis
        )

        self.load_df(df)

    def add_cats(self, axis, cat_data):
        """
        Add categories to rows or columns using cat_data array of objects. Each object in cat_data is a dictionary with one key (category title) and value (rows/column names) that have this category. Categories will be added onto the existing categories and will be added in the order of the objects in the array.

        Example ``cat_data``::


            [
              {
                "title": "First Category",
                "cats": {
                  "true": [
                    "ROS1",
                    "AAK1"
                  ]
                }
              },
              {
                "title": "Second Category",
                "cats": {
                  "something": [
                    "PDK4"
                  ]
                }
              }
            ]


        """
        for inst_data in cat_data:
            categories.add_cats(self, axis, inst_data)

    def iframe_web_app(self, filename=None, width=1000, height=800):
        return iframe_web_app.main(self, filename, width, height)

    def enrichrgram(self, lib, axis="row"):
        """
        Add Enrichr gene enrichment results to your visualization (where your rows
        are genes). Run enrichrgram before clustering to incldue enrichment results
        as row categories. Enrichrgram can also be run on the front-end using the
        Enrichr logo at the top left.

        Set lib to the Enrichr library that you want to use for enrichment analysis.
        Libraries included:

          * ChEA_2016
          * KEA_2015
          * ENCODE_TF_ChIP-seq_2015
          * ENCODE_Histone_Modifications_2015
          * Disease_Perturbations_from_GEO_up
          * Disease_Perturbations_from_GEO_down
          * GO_Molecular_Function_2015
          * GO_Biological_Process_2015
          * GO_Cellular_Component_2015
          * Reactome_2016
          * KEGG_2016
          * MGI_Mammalian_Phenotype_Level_4
          * LINCS_L1000_Chem_Pert_up
          * LINCS_L1000_Chem_Pert_down

        """

        df = self.export_df()
        df, bar_info = enr_fun.add_enrichr_cats(df, axis, lib)
        self.load_df(df)

        self.dat["enrichrgram_lib"] = lib
        self.dat["row_cat_bars"] = bar_info

    @staticmethod
    def load_gmt(filename):
        return load_data.load_gmt(filename)

    @staticmethod
    def load_json_to_dict(filename):
        return load_data.load_json_to_dict(filename)

    @staticmethod
    def save_dict_to_json(inst_dict, filename, indent="no-indent"):
        export_data.save_dict_to_json(inst_dict, filename, indent)

    @staticmethod
    def load_gene_exp_to_df(inst_path):
        """
        Loads gene expression data from 10x in sparse matrix format and returns a
        Pandas dataframe
        """

        from ast import literal_eval as make_tuple

        import pandas as pd
        from scipy import io

        # matrix
        matrix = io.mmread(f"{inst_path}matrix.mtx")
        mat = matrix.todense()

        # genes
        filename = f"{inst_path}genes.tsv"
        with Path.open(filename) as f:
            lines = f.readlines()

        # # add unique id to all genes
        # genes = []
        # unique_id = 0
        # for inst_line in lines:
        #     inst_line = inst_line.strip().split()

        #     if len(inst_line) > 1:
        #       inst_gene = inst_line[1]
        #     else:
        #       inst_gene = inst_line[0]

        #     genes.append(inst_gene + '_' + str(unique_id))
        #     unique_id = unique_id + 1

        # add unique id only to duplicate genes
        ini_genes = []
        for inst_line in lines:
            inst_line = inst_line.strip().split()
            inst_gene = inst_line[1] if len(inst_line) > 1 else inst_line[0]
            ini_genes.append(inst_gene)

        gene_name_count = pd.Series(ini_genes).value_counts()
        duplicate_genes = gene_name_count[gene_name_count > 1].index.tolist()

        dup_index = {}
        genes = []
        for inst_row in ini_genes:
            # add index to non-unique genes
            if inst_row in duplicate_genes:
                # calc_non-unque index
                dup_index[inst_row] = dup_index.get(inst_row, 0) + 1

                new_row = f"{inst_row}_{dup_index[inst_row]}"

            else:
                new_row = inst_row

            genes.append(new_row)

        # barcodes
        filename = f"{inst_path}barcodes.tsv"
        with Path.open(filename) as f:
            lines = f.readlines()

        cell_barcodes = []
        for inst_bc in lines:
            inst_bc = inst_bc.strip().split("\t")

            # remove dash from barcodes if necessary
            if "-" in inst_bc[0]:
                inst_bc[0] = inst_bc[0].split("-")[0]

            cell_barcodes.append(inst_bc[0])

        # parse tuples if necessary
        with contextlib.suppress(Exception):
            cell_barcodes = [make_tuple(x) for x in cell_barcodes]

        with contextlib.suppress(Exception):
            genes = [make_tuple(x) for x in genes]

        # make dataframe
        return pd.DataFrame(mat, index=genes, columns=cell_barcodes)

    @staticmethod
    def save_gene_exp_to_mtx_dir(inst_path, df):
        from scipy import io, sparse

        if not Path.exists(inst_path):
            Path.makedirs(inst_path)

        genes = df.index.tolist()
        barcodes = df.columns.tolist()

        Network.save_list_to_tsv(genes, f"{inst_path}genes.tsv")
        Network.save_list_to_tsv(barcodes, f"{inst_path}barcodes.tsv")

        mat_ge = df.values
        mat_ge_sparse = sparse.coo_matrix(mat_ge)

        io.mmwrite(f"{inst_path}matrix.mtx", mat_ge_sparse)

    @staticmethod
    def save_list_to_tsv(inst_list, filename):
        with Path.open(filename, "w") as f:
            for inst_line in inst_list:
                f.write(f"{inst_line}\n")

    def sim_same_and_diff_category_samples(
        self,
        df,
        cat_index=1,
        dist_type="cosine",
        equal_var=False,
        plot_roc=True,
        precalc_dist=False,
        calc_roc=True,
    ):
        """
        Calculate the similarity of samples from the same and different categories. The
        cat_index gives the index of the category, where 1 in the first category
        """

        cols = df.columns.tolist()

        if isinstance(precalc_dist, bool):
            # compute distnace between rows (transpose to get cols as rows)
            dist_arr = 1 - pdist(df.transpose(), metric=dist_type)
        else:
            dist_arr = precalc_dist

        # generate sample names with categories
        sample_combos = list(combinations(range(df.shape[1]), 2))

        sample_names = [
            f"{ind}_same" if cols[x[0]][cat_index] == cols[x[1]][cat_index] else f"{ind}_different"
            for ind, x in enumerate(sample_combos)
        ]

        ser_dist = pd.Series(data=dist_arr, index=sample_names)

        # find same-cat sample comparisons
        same_cat = [x for x in sample_names if x.split("_")[1] == "same"]

        # find diff-cat sample comparisons
        diff_cat = [x for x in sample_names if x.split("_")[1] == "different"]

        # make series of same and diff category sample comparisons
        ser_same = ser_dist[same_cat]
        ser_same.name = "Same Category"
        ser_diff = ser_dist[diff_cat]
        ser_diff.name = "Different Category"

        sim_dict = {"same": ser_same, "diff": ser_diff}
        roc_data = {}
        pval_dict = {}

        _, pval_dict["ttest"] = ttest_ind(ser_diff, ser_same, equal_var=equal_var)

        _, pval_dict["mannwhitney"] = mannwhitneyu(ser_diff, ser_same)

        if calc_roc:
            # calc AUC
            true_index = list(np.ones(sim_dict["same"].shape[0]))
            false_index = list(np.zeros(sim_dict["diff"].shape[0]))
            y_true = true_index + false_index

            true_val = list(sim_dict["same"].values)
            false_val = list(sim_dict["diff"].values)
            y_score = true_val + false_val

            fpr, tpr, thresholds = roc_curve(y_true, y_score)

            inst_auc = auc(fpr, tpr)

            if plot_roc:
                plt.figure()
                plt.plot(fpr, tpr)
                plt.plot([0, 1], [0, 1], color="navy", linestyle="--")
                plt.figure(figsize=(10, 10))

                print("AUC", inst_auc)

            roc_data = {
                "true": y_true,
                "score": y_score,
                "fpr": fpr,
                "tpr": tpr,
                "thresholds": thresholds,
                "auc": inst_auc,
            }

        return {
            "sim_dict": sim_dict,
            "pval_dict": pval_dict,
            "roc_data": roc_data,
        }

    def generate_signatures(
        self,
        df_data,
        df_meta,
        category_name,
        pval_cutoff=0.05,
        num_top_dims=False,
        verbose=True,
        equal_var=False,
    ):
        """
        Generate signatures for column categories
        T-test is run for each category in a one-vs-all manner. The num_top_dims overrides
        the P-value cutoff.
        """

        df_t = df_data.transpose()

        # remove columns (dimensions) with constant values
        orig_num_cols = df_t.shape[1]
        df_t = df_t.loc[:, (df_t != df_t.iloc[0]).any()]
        if df_t.shape[1] < orig_num_cols:
            print("dropped columns with constant values")

        # make tuple rows
        df_t.index = [(x, df_meta.loc[x, category_name]) for x in df_t.index.tolist()]
        category_level = 1

        df = self.row_tuple_to_multiindex(df_t)

        cell_types = sorted(set(df.index.get_level_values(category_level).tolist()))

        keep_genes = []
        keep_genes_dict = {}
        gene_pval_dict = {}
        all_fold_info = {}

        for inst_ct in cell_types:
            inst_ct_mat = df.xs(key=inst_ct, level=category_level)
            inst_other_mat = df.drop(inst_ct, level=category_level)

            # save mean values and fold change
            fold_info = {
                "cluster_mean": inst_ct_mat.mean(),
                "other_mean": inst_other_mat.mean(),
            }
            fold_info["log2_fold"] = fold_info["cluster_mean"] / fold_info["other_mean"]
            fold_info["log2_fold"] = fold_info["log2_fold"].apply(np.log2)
            all_fold_info[inst_ct] = fold_info

            _, inst_pvals = ttest_ind(inst_ct_mat, inst_other_mat, axis=0, equal_var=equal_var)

            ser_pval = pd.Series(data=inst_pvals, index=df.columns.tolist()).sort_values()

            ser_pval_keep = (
                ser_pval[:num_top_dims] if num_top_dims else ser_pval[ser_pval < pval_cutoff]
            )

            gene_pval_dict[inst_ct] = ser_pval_keep

            inst_keep = ser_pval_keep.index.tolist()
            keep_genes.extend(inst_keep)
            keep_genes_dict[inst_ct] = inst_keep

        keep_genes = sorted(set(keep_genes))

        df_gbm = df.groupby(level=category_level).mean().transpose()

        df_sig = df_gbm.loc[keep_genes]

        if len(keep_genes) == 0 and verbose:
            print("found no informative dimensions")

        df_gene_pval = pd.concat(gene_pval_dict, axis=1, sort=False)

        df_diff = {}
        for inst_col in df_gene_pval:
            inst_pvals = df_gene_pval[inst_col]

            # nans represent dimensions that did not meet pval or top threshold
            inst_pvals = inst_pvals.dropna().sort_values()

            inst_genes = inst_pvals.index.tolist()

            # prevent failure if no cells have this category
            if inst_pvals.shape[0] > 0:
                _, pval_corr = smm.multipletests(inst_pvals, 0.05, method="fdr_bh")[:2]

                ser_pval = pd.Series(data=inst_pvals, index=inst_genes, name="P-values")
                ser_bh_pval = pd.Series(data=pval_corr, index=inst_genes, name="BH P-values")
                ser_log2_fold = all_fold_info[inst_col]["log2_fold"].loc[inst_genes]
                ser_log2_fold.name = "Log2 Fold Change"
                ser_cluster_mean = all_fold_info[inst_col]["cluster_mean"].loc[inst_genes]
                ser_cluster_mean.name = "Cluster Mean"
                ser_other_mean = all_fold_info[inst_col]["other_mean"].loc[inst_genes]
                ser_other_mean.name = "All Other Mean"
                inst_df = pd.concat(
                    [ser_pval, ser_bh_pval, ser_log2_fold, ser_cluster_mean, ser_other_mean], axis=1
                )

                df_diff[inst_col] = inst_df

        return df_sig, df_diff

    def old_generate_signatures(
        self,
        df_ini,
        category_level,
        pval_cutoff=0.05,
        num_top_dims=False,
        verbose=True,
        equal_var=False,
    ):
        """Generate signatures for column categories"""

        # change this to use category name and set caetgory level to 1 always
        ###################################

        df_t = df_ini.transpose()

        # remove columns with constant values
        df_t = df_t.loc[:, (df_t != df_t.iloc[0]).any()]

        df = self.row_tuple_to_multiindex(df_t)

        cell_types = sorted(set(df.index.get_level_values(category_level).tolist()))

        keep_genes = []
        keep_genes_dict = {}
        gene_pval_dict = {}
        all_fold_info = {}

        for inst_ct in cell_types:
            inst_ct_mat = df.xs(key=inst_ct, level=category_level)
            inst_other_mat = df.drop(inst_ct, level=category_level)

            # save mean values and fold change
            fold_info = {
                "cluster_mean": inst_ct_mat.mean(),
                "other_mean": inst_other_mat.mean(),
            }
            fold_info["log2_fold"] = fold_info["cluster_mean"] / fold_info["other_mean"]
            fold_info["log2_fold"] = fold_info["log2_fold"].apply(np.log2)
            all_fold_info[inst_ct] = fold_info

            _, inst_pvals = ttest_ind(inst_ct_mat, inst_other_mat, axis=0, equal_var=equal_var)

            ser_pval = pd.Series(data=inst_pvals, index=df.columns.tolist()).sort_values()

            ser_pval_keep = (
                ser_pval[:num_top_dims] if num_top_dims else ser_pval[ser_pval < pval_cutoff]
            )

            gene_pval_dict[inst_ct] = ser_pval_keep

            inst_keep = ser_pval_keep.index.tolist()
            keep_genes.extend(inst_keep)
            keep_genes_dict[inst_ct] = inst_keep

        keep_genes = sorted(set(keep_genes))

        df_gbm = df.groupby(level=category_level).mean().transpose()
        cols = df_gbm.columns.tolist()
        new_cols = [(inst_col, f"{category_level + 1}: {inst_col}") for inst_col in cols]
        df_gbm.columns = new_cols

        df_sig = df_gbm.loc[keep_genes]

        if len(keep_genes) == 0 and verbose:
            print("found no informative dimensions")

        df_gene_pval = pd.concat(gene_pval_dict, axis=1, sort=False)

        return df_sig, df_gene_pval, all_fold_info

    def predict_cats_from_sigs(
        self,
        df_data_ini,
        df_meta,
        df_sig_ini,
        predict="Predicted Category",
        dist_type="cosine",
        unknown_thresh=-1,
    ):
        """Predict category using signature"""

        keep_rows = df_sig_ini.index.tolist()
        data_rows = df_data_ini.index.tolist()

        common_rows = list(set(data_rows).intersection(keep_rows))

        df_data = deepcopy(df_data_ini.loc[common_rows])
        df_sig = deepcopy(df_sig_ini.loc[common_rows])

        # calculate sim_mat of df_data and df_sig
        cell_types = df_sig.columns.tolist()
        barcodes = df_data.columns.tolist()
        sim_mat = 1 - pairwise_distances(df_sig.transpose(), df_data.transpose(), metric=dist_type)
        df_sim = pd.DataFrame(data=sim_mat, index=cell_types, columns=barcodes).transpose()

        # get the top column value (most similar signature)
        df_sim_top = df_sim.idxmax(axis=1)

        # get the maximum similarity of a cell to a cell type definition
        max_sim = df_sim.max(axis=1)

        unknown_cells = max_sim[max_sim < unknown_thresh].index.tolist()

        # assign unknown cells (need category of same name)
        df_sim_top[unknown_cells] = "Unknown"

        # add predicted category name to top list
        # top_list = df_sim_top.values
        df_sim = df_sim.transpose()

        df_meta[predict] = df_sim_top

        return df_sim, df_meta

    def old_predict_cats_from_sigs(
        self,
        df_data_ini,
        df_sig_ini,
        dist_type="cosine",
        predict_level="Predict Category",
        truth_level=1,
        unknown_thresh=-1,
    ):
        """Predict category using signature"""

        keep_rows = df_sig_ini.index.tolist()
        data_rows = df_data_ini.index.tolist()

        common_rows = list(set(data_rows).intersection(keep_rows))

        df_data = deepcopy(df_data_ini.loc[common_rows])
        df_sig = deepcopy(df_sig_ini.loc[common_rows])

        # calculate sim_mat of df_data and df_sig
        cell_types = df_sig.columns.tolist()
        barcodes = df_data.columns.tolist()
        sim_mat = 1 - pairwise_distances(df_sig.transpose(), df_data.transpose(), metric=dist_type)
        df_sim = pd.DataFrame(data=sim_mat, index=cell_types, columns=barcodes).transpose()

        # get the top column value (most similar signature)
        df_sim_top = df_sim.idxmax(axis=1)

        # get the maximum similarity of a cell to a cell type definition
        max_sim = df_sim.max(axis=1)

        unknown_cells = max_sim[max_sim < unknown_thresh].index.tolist()

        # assign unknown cells (need category of same name)
        df_sim_top[unknown_cells] = "Unknown"

        # add predicted category name to top list
        top_list = df_sim_top.values
        top_list = [f"{predict_level}: {x[0] if isinstance(x, tuple) else x}" for x in top_list]

        # add cell type category to input data
        df_cat = deepcopy(df_data)
        cols = df_cat.columns.tolist()

        # check whether the columns have the true category available
        has_truth = isinstance(cols[0], tuple)

        new_cols = (
            [(*a, b) for a, b in zip(cols, top_list, strict=False)]
            if has_truth
            else list(zip(cols, top_list, strict=False))
        )

        # transfer new categories
        df_cat.columns = new_cols

        # keep track of true and predicted labels
        y_info = {"true": [], "pred": []}

        if has_truth:
            y_info["true"] = [x[truth_level].split(": ")[1] for x in cols]
            y_info["pred"] = [x.split(": ")[1] for x in top_list]

        return df_cat, df_sim.transpose(), y_info

    def assess_prediction(self, df_meta, truth, pred):
        """Generate confusion matrix from y_info"""

        y_info = {
            "true": df_meta[truth].values.tolist(),
            "pred": df_meta[pred].values.tolist(),
        }

        sorted_cats = sorted(set(y_info["true"] + y_info["pred"]))
        conf_mat = confusion_matrix(y_info["true"], y_info["pred"], labels=sorted_cats)

        # true columns and pred rows
        df_conf = pd.DataFrame(conf_mat, index=sorted_cats, columns=sorted_cats).transpose()

        total_correct = np.trace(df_conf)
        total_pred = df_conf.sum().sum()
        fraction_correct = total_correct / float(total_pred)

        # calculate ser_correct
        correct_list = []
        cat_counts = df_conf.sum(axis=0)
        all_cols = df_conf.columns.tolist()
        for inst_cat in all_cols:
            inst_correct = df_conf.loc[inst_cat, inst_cat] / cat_counts[inst_cat]
            correct_list.append(inst_correct)

        ser_correct = pd.Series(data=correct_list, index=all_cols)

        return df_conf, ser_correct, fraction_correct

    def old_confusion_matrix_and_correct_series(self, y_info):
        """Generate confusion matrix from y_info"""

        a = deepcopy(y_info["true"])
        true_count = {i: a.count(i) for i in set(a)}

        a = deepcopy(y_info["pred"])
        pred_count = {i: a.count(i) for i in set(a)}

        sorted_cats = sorted(set(y_info["true"] + y_info["pred"]))
        conf_mat = confusion_matrix(y_info["true"], y_info["pred"], sorted_cats)
        df_conf = pd.DataFrame(conf_mat, index=sorted_cats, columns=sorted_cats)

        total_correct = np.trace(df_conf)
        total_pred = df_conf.sum().sum()
        fraction_correct = total_correct / float(total_pred)

        # calculate ser_correct
        correct_list = []
        cat_counts = df_conf.sum(axis=1)
        all_cols = df_conf.columns.tolist()
        for inst_cat in all_cols:
            inst_correct = df_conf[inst_cat].loc[inst_cat] / cat_counts[inst_cat]
            correct_list.append(inst_correct)

        ser_correct = pd.Series(data=correct_list, index=all_cols)

        populations = {"true": true_count, "pred": pred_count}

        return df_conf, populations, ser_correct, fraction_correct

    def compare_performance_to_shuffled_labels(
        self,
        df_data,
        category_level,
        num_shuffles=100,
        random_seed=99,
        pval_cutoff=0.05,
        dist_type="cosine",
        num_top_dims=False,
        predict_level="Predict Category",
        truth_level=1,
        unknown_thresh=-1,
        equal_var=False,
        performance_type="prediction",
    ):
        random.seed(random_seed)

        perform_list = []

        # pre-calculate the distance matrix (similarity matrix) if necessary
        if performance_type == "cat_sim_auc":
            dist_arr = 1 - pdist(df_data.transpose(), metric=dist_type)

        for inst_run in range(num_shuffles + 1):
            cols = df_data.columns.tolist()
            rows = df_data.index.tolist()
            mat = df_data.values

            shuffled_cols = deepcopy(cols)
            random.shuffle(shuffled_cols)

            # do not perform shuffling the first time to confirm that we get the same
            # results as the unshuffled dataaset
            df_shuffle = (
                deepcopy(df_data)
                if inst_run == 0
                else pd.DataFrame(data=mat, columns=shuffled_cols, index=rows)
            )

            # generate signature on shuffled data
            df_sig, keep_genes, keep_genes_dict, fold_info = self.generate_signatures(
                df_shuffle,
                category_level,
                pval_cutoff=pval_cutoff,
                num_top_dims=num_top_dims,
                equal_var=equal_var,
            )

            # predictive performance
            if performance_type == "prediction":
                # predict categories from signature
                _, _, y_info = self.predict_cats_from_sigs(
                    df_shuffle,
                    df_sig,
                    dist_type=dist_type,
                    predict_level=predict_level,
                    truth_level=truth_level,
                    unknown_thresh=unknown_thresh,
                )

                # calc confusion matrix and performance
                _, _, _, fraction_correct = self.confusion_matrix_and_correct_series(y_info)

                # store performances of shuffles
                if inst_run > 0:
                    perform_list.append(fraction_correct)
                else:
                    real_performance = fraction_correct
                    print(f"performance (fraction correct) of unshuffled: {real_performance}")

            elif performance_type == "cat_sim_auc":
                # predict categories from signature
                sim_data = self.sim_same_and_diff_category_samples(
                    df_shuffle,
                    cat_index=1,
                    plot_roc=False,
                    equal_var=equal_var,
                    precalc_dist=dist_arr,
                )

                # store performances of shuffles
                if inst_run > 0:
                    perform_list.append(sim_data["roc_data"]["auc"])
                else:
                    real_performance = sim_data["roc_data"]["auc"]
                    print(
                        f"performance (category similarity auc) of unshuffled: {sim_data['roc_data']['auc']}"
                    )

        perform_ser = pd.Series(perform_list)

        in_top_fraction = perform_ser[perform_ser > real_performance].shape[0] / num_shuffles
        print(f"real data performs in the top {in_top_fraction * 100}% of shuffled labels\n")

        return perform_ser

    @staticmethod
    def box_scatter_plot(
        df,
        group,
        columns=False,
        rand_seed=100,
        alpha=0.5,
        dot_color="red",
        num_row=None,
        num_col=1,
        figsize=(10, 10),
        start_title="Variable Measurements Across",
        end_title="Groups",
        group_list=False,
    ):
        import matplotlib.pyplot as plt
        import pandas as pd
        from scipy import stats
        # %matplotlib inline

        if not columns:
            columns = df.columns.tolist()

        plt.figure(figsize=figsize)
        figure_title = f"{start_title} {group} {end_title}"
        plt.suptitle(figure_title, fontsize=20)

        # list of arranged dataframes
        dfs = {}

        for col_num, column in enumerate(columns):
            plot_id = col_num + 1

            # group by column name or multiIndex name
            grouped = df.groupby(group) if group in df.columns.tolist() else df.groupby(level=group)

            names, vals = [], []

            column_title = column[0] if isinstance(column, tuple) else column

            for name, subdf in grouped:
                names.append(name)

                inst_ser = subdf[column]

                column_name = f"{column_title}-{name}"

                inst_ser.name = column_name
                vals.append(inst_ser)

                np.random.seed(rand_seed)

            ax = plt.subplot(num_row, num_col, plot_id)

            plt.boxplot(vals, labels=names)

            ngroup = len(vals)

            clevels = np.linspace(0.0, 1.0, ngroup)

            for clevel in clevels:
                xs = np.random.normal(
                    clevels.tolist().index(clevel) + 1,
                    0.04,
                    len(vals[clevels.tolist().index(clevel)]),
                )
                plt.subplot(num_row, num_col, plot_id)
                plt.scatter(xs, vals[clevels.tolist().index(clevel)], c=dot_color, alpha=alpha)

            df_arranged = pd.concat(vals, axis=1)

            # anova
            anova_data = [df_arranged[col].dropna() for col in df_arranged]
            f_val, pval = stats.f_oneway(*anova_data)

            title_text = (
                f"{column_title} P-val: {pval:.2e}"
                if pval < 0.01
                else f"{column_title} P-val: {round(pval * 100000) / 100000}"
            )
            ax.set_title(title_text)

            dfs[column] = df_arranged

        return dfs

    @staticmethod
    def rank_cols_by_anova_pval(
        df,
        group,
        columns=False,
        rand_seed=100,
        alpha=0.5,
        dot_color="red",
        num_row=None,
        num_col=1,
        figsize=(10, 10),
    ):
        import numpy as np
        import pandas as pd
        from scipy import stats

        # import matplotlib.pyplot as plt
        # %matplotlib inline

        if not columns:
            columns = df.columns.tolist()

        # plt.figure(figsize=figsize)

        # list of arranged dataframes
        pval_list = []

        for column in columns:
            # group by column name or multiIndex name
            grouped = df.groupby(group) if group in df.columns.tolist() else df.groupby(level=group)

            names, vals = [], []

            column_title = column[0] if isinstance(column, tuple) else column

            for name, subdf in grouped:
                names.append(name)

                inst_ser = subdf[column]

                column_name = f"{column_title}-{name}"

                inst_ser.name = column_name
                vals.append(inst_ser)

                np.random.seed(rand_seed)

            df_arranged = pd.concat(vals, axis=1)

            # anova
            anova_data = [df_arranged[col].dropna() for col in df_arranged]
            _, pval = stats.f_oneway(*anova_data)

            pval_list.append(pval)

        pval_ser = pd.Series(data=pval_list, index=columns)
        return pval_ser.sort_values(ascending=True)

    def row_tuple_to_multiindex(self, df):
        from copy import deepcopy

        import pandas as pd

        df_mi = deepcopy(df)
        rows = df_mi.index.tolist()
        titles = []
        for inst_part in rows[0]:
            inst_title = inst_part.split(": ")[0] if ": " in inst_part else "Name"
            titles.append(inst_title)

        new_rows = []
        for inst_row in rows:
            inst_row = list(inst_row)
            new_row = []
            for inst_part in inst_row:
                inst_part = inst_part.split(": ")[1] if ": " in inst_part else inst_part
                new_row.append(inst_part)
            new_row = tuple(new_row)
            new_rows.append(new_row)

        df_mi.index = new_rows

        df_mi.index = pd.MultiIndex.from_tuples(df_mi.index, names=titles)

        return df_mi

    def set_cat_colors(self, cat_colors, axis, cat_index, cat_title=False):
        for inst_ct in cat_colors:
            cat_name = f"{cat_title}: {inst_ct}" if cat_title else inst_ct

            inst_color = cat_colors[inst_ct]
            self.set_cat_color(
                axis=axis, cat_index=cat_index, cat_name=cat_name, inst_color=inst_color
            )

    def set_manual_category(self, col=None, row=None, preferred_cats=None):
        """
        This method is used to tell Clustergrammer2 that the user wants to define
        a manual category interactively using the dendrogram.
        """

        self.dat["manual_category"] = {"col": col, "row": row}

        if preferred_cats is not None:
            pref_cats = []
            for inst_row in preferred_cats.index.tolist():
                inst_dict = {
                    "name": inst_row,
                    "color": preferred_cats.loc[inst_row, "color"],
                }
                pref_cats.append(inst_dict)

            if col is not None:
                self.dat["manual_category"]["col_cats"] = pref_cats

            if row is not None:
                self.dat["manual_category"]["row_cats"] = pref_cats

    def ds_to_original_meta(self, axis):
        clusters = self.meta_ds_col.index.tolist()

        man_cat_title = self.dat["manual_category"][axis]

        for inst_cluster in clusters:
            # get the manual category from the downsampled data
            inst_man_cat = self.meta_ds_col.loc[inst_cluster, man_cat_title]

            # find original labels that are assigned to this cluster
            found_labels = self.meta_col[self.meta_col[self.ds_name] == inst_cluster].index.tolist()

            # update manual category
            self.meta_col.loc[found_labels, man_cat_title] = inst_man_cat

    def get_manual_category(self, tmp):
        for axis in ["col"]:
            with contextlib.suppress(Exception):
                ###########################################################

                export_dict = {}

                inst_nodes = self.dat["nodes"][axis]
                # print('get get_manual_category', len(inst_nodes))

                cat_title = self.dat["manual_category"][axis]

                # Category Names
                with contextlib.suppress(Exception):
                    ###########################################################

                    export_dict[cat_title] = pd.Series(
                        json.loads(self.widget_instance.manual_cat)[axis][cat_title]
                    )

                    if hasattr(self, "meta_cat"):
                        if axis == "row":
                            if not self.is_downsampled:
                                self.meta_row.loc[inst_nodes, cat_title] = export_dict[cat_title]
                            else:
                                self.meta_ds_row.loc[inst_nodes, cat_title] = export_dict[cat_title]
                                self.ds_to_original_meta(axis)

                        elif axis == "col":
                            if not self.is_downsampled:
                                # print('> > > > > > > > > > > ')
                                # print(export_dict[cat_title])

                                # print('.. .. .. .. .. .. .. .. .. .. .. .. ')
                                # print(self.meta_col)

                                # # for some reason this is breaking when trying to sync metadata
                                # # switching to slower row based syncing
                                # self.meta_col.loc[inst_nodes, cat_title] = export_dict[cat_title]

                                # manually looping through rows in metadata works
                                for inst_row in export_dict[cat_title].index.tolist():
                                    self.meta_col.loc[inst_row, cat_title] = export_dict[cat_title][
                                        inst_row
                                    ]

                                # print('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>')
                                # print(self.meta_col)

                            else:
                                self.meta_ds_col.loc[inst_nodes, cat_title] = export_dict[cat_title]
                                self.ds_to_original_meta(axis)

                # Category Colors
                #######################
                export_dict["cat_colors"] = json.loads(self.widget_instance.manual_cat)[
                    "global_cat_colors"
                ]

                # # drop title from category colors
                # export_dict['cat_colors'] = {}
                # for inst_cat in ini_new_colors:
                #   if (': ' in inst_cat):
                #     export_dict['cat_colors'][inst_cat.split(': ')[1]] = ini_new_colors[inst_cat]
                #   else:
                #     export_dict['cat_colors'][inst_cat] = ini_new_colors[inst_cat]

                # if hasattr(self, 'meta_cat') == False:
                #   return export_dict

    @staticmethod
    def umi_norm(df):
        # umi norm
        barcode_umi_sum = df.sum()
        return df.div(barcode_umi_sum)

    @staticmethod
    def make_df_from_cols(cols):
        inst_col = cols[0]

        cat_titles = []
        for inst_info in inst_col[1:]:
            inst_title = inst_info.split(": ")[0]
            cat_titles.append(inst_title)

        clean_cols = []
        for inst_col in cols:
            inst_clean = [info.split(": ", 1)[1] if ": " in info else info for info in inst_col]
            clean_cols.append(tuple(inst_clean))

        df_ini = pd.DataFrame(data=clean_cols).set_index(0)
        mat = df_ini.values
        rows = df_ini.index.tolist()

        return pd.DataFrame(data=mat, index=rows, columns=cat_titles)


def save_list_to_tsv(inst_list, filename):
    with Path.open(filename, "w") as f:
        for inst_line in inst_list:
            f.write(f"{inst_line}\n")


def generate_signatures(
    df_shuffle,
    category_level,
    pval_cutoff=0.05,
    num_top_dims=False,
    equal_var=False,
):
    """Generate signatures for column categories - standalone function"""
    # This is a placeholder implementation - the actual implementation
    # would need to be provided based on the original function
    return None, [], {}, {}



from . import calc_clust, enrichr_functions as enr_fun, make_sim_mat


def make_clust(
    net,
    dist_type="cosine",
    run_clustering=True,
    dendro=True,
    requested_views=None,
    linkage_type="average",
    sim_mat=False,
    filter_sim=0.0,
    calc_cat_pval=False,
    sim_mat_views=None,
    run_enrichr=None,
    enrichrgram=None,
    clust_library="scipy",
    min_samples=1,
    min_cluster_size=2,
):
    """
    This will perform hierarchical clustering
    """
    if requested_views is None:
        requested_views = ["pct_row_sum", "N_row_sum"]

    if sim_mat_views is None:
        sim_mat_views = ["N_row_sum"]

    # threshold = 0.0001
    # df = run_filter.df_filter_row_sum(df, threshold)
    # df = run_filter.df_filter_col_sum(df, threshold)

    if run_enrichr is not None:
        df = net.dat_to_df()
        df = enr_fun.add_enrichr_cats(df, "row", run_enrichr)
        net.df_to_dat(df, define_cat_colors=True)

    inst_dm = calc_clust.cluster_row_and_col(
        net,
        dist_type=dist_type,
        linkage_type=linkage_type,
        run_clustering=run_clustering,
        dendro=dendro,
        ignore_cat=False,
        calc_cat_pval=calc_cat_pval,
        clust_library=clust_library,
        min_samples=min_samples,
        min_cluster_size=min_cluster_size,
    )

    which_sim = []

    if sim_mat is True:
        which_sim = ["row", "col"]
    elif sim_mat == "row":
        which_sim = ["row"]
    elif sim_mat == "col":
        which_sim = ["col"]

    if sim_mat is not False:
        sim_net = make_sim_mat.main(net, inst_dm, which_sim, filter_sim, sim_mat_views)

        net.sim = {}

        for inst_rc in which_sim:
            net.sim[inst_rc] = sim_net[inst_rc].viz

            other_rc = "col" if inst_rc == "row" else "row"

            # keep track of cat_colors
            net.sim[inst_rc]["cat_colors"][inst_rc] = net.viz["cat_colors"][inst_rc]
            net.sim[inst_rc]["cat_colors"][other_rc] = net.viz["cat_colors"][inst_rc]

    else:
        net.sim = {}

    net.viz["views"] = []

    if enrichrgram is not None:
        # toggle enrichrgram functionality from back-end
        net.viz["enrichrgram"] = enrichrgram

    if "enrichrgram_lib" in net.dat:
        net.viz["enrichrgram"] = True
        net.viz["enrichrgram_lib"] = net.dat["enrichrgram_lib"]

    if "row_cat_bars" in net.dat:
        net.viz["row_cat_bars"] = net.dat["row_cat_bars"]



def main(self, widget=None):
    if not hasattr(self, "meta_cat"):
        self.meta_cat = False

    self.dat = {
        "nodes": {"row": [], "col": []},
        "mat": [],
        "node_info": {},
    }

    for inst_rc in self.dat["nodes"]:
        self.dat["node_info"][inst_rc] = {
            "ini": [],
            "clust": [],
            "rank": [],
            "info": [],
            "cat": [],
            "value": [],
        }

    # check if net has categories predefined
    if not hasattr(self, "persistent_cat_colors"):
        self.persistent_cat_colors = False
        found_cats = False
    else:
        found_cats = True
        inst_cat_colors = self.viz["cat_colors"]
        inst_global_cat_colors = self.viz["global_cat_colors"]

    # initialize matrix colors
    ###########################
    has_matrix_colors = hasattr(self, "viz") and "matrix_colors" in self.viz

    matrix_colors = (
        self.viz["matrix_colors"] if has_matrix_colors else {"pos": "red", "neg": "blue"}
    )

    # add widget if necessary
    if widget is not None:
        self.widget_class = widget

    self.is_downsampled = False

    self.viz = {
        "row_nodes": [],
        "col_nodes": [],
        "links": [],
        "mat": [],
        "matrix_colors": matrix_colors,
    }

    if not found_cats:
        self.viz["cat_colors"] = {"row": {}, "col": {}}
        self.viz["global_cat_colors"] = {}
    else:
        self.viz["cat_colors"] = inst_cat_colors
        self.viz["global_cat_colors"] = inst_global_cat_colors

    self.sim = {}
    self.umap = {}


def viz(self, reset_cat_colors=False):
    # keep track of old cat_colors
    old_cat_colors = self.viz["cat_colors"]
    old_global_cat_colors = self.viz["global_cat_colors"]

    matrix_colors = self.viz.get("matrix_colors", {"pos": "red", "neg": "blue"})

    self.viz = {
        "row_nodes": [],
        "col_nodes": [],
        "links": [],
        "mat": [],
        "matrix_colors": matrix_colors,
    }

    if reset_cat_colors:
        self.viz["cat_colors"] = {"row": {}, "col": {}}
        self.viz["global_cat_colors"] = {}
    else:
        self.viz["cat_colors"] = old_cat_colors
        self.viz["global_cat_colors"] = old_global_cat_colors



"""
Module for visualization
"""

from ipywidgets import HBox, Layout, jslink

from .local_server import get_local_server
from .widget import Landscape, Matrix


def landscape_matrix(landscape, mat, width="600px", height="700px"):
    """
    Display a `Landscape` widget and a `Matrix` widget side by side.

    Args:
        landscape (Landscape): A `Landscape` widget.
        mat (Matrix): A `Matrix` widget.
        width (str): The width of the widgets.
        height (str): The height of the widgets.

    Returns:
        HBox: Visualization display containing both widgets

    Example:
    See example [Landscape-Matrix_Xenium](../../../examples/brief_notebooks/Landscape-Matrix_Xenium) notebook
    """
    # Use `jslink` to directly link `click_info` from `mat` to `trigger_value` in `landscape_ist`
    jslink((mat, "click_info"), (landscape, "update_trigger"))

    # Set layouts for the widgets
    mat.layout = Layout(width=width)  # Adjust as needed
    landscape.layout = Layout(width=width, height=height)  # Adjust as needed

    return HBox([landscape, mat])


__all__ = ["Landscape", "Matrix", "get_local_server", "landscape_matrix"]



def main(df):
    """
    1) check that rows are strings (in case of numerical names)
    2) check for tuples, and in that case load tuples to categories
    """
    from ast import literal_eval as make_tuple

    import numpy as np

    test = {
        "row": df.index.tolist(),
        "col": df.columns.tolist(),
    }

    # if type( test_row ) is not str and type( test_row ) is not tuple:

    found_tuple = {}
    found_number = {}
    for inst_rc in ["row", "col"]:
        inst_name = test[inst_rc][0]

        found_tuple[inst_rc] = False
        found_number[inst_rc] = False

        if not isinstance(inst_name, tuple):
            if isinstance(inst_name, (int | float | np.int64)):
                found_number[inst_rc] = True

            else:
                check_open = inst_name[0]
                check_comma = inst_name.find(",")
                check_close = inst_name[-1]

                if (
                    check_open == "("
                    and check_close == ")"
                    and check_comma > 0
                    and check_comma < len(inst_name)
                ):
                    found_tuple[inst_rc] = True

    # convert to tuple if necessary
    #################################################
    if found_tuple["row"]:
        row_names = df.index.tolist()
        row_names = [make_tuple(x) for x in row_names]
        df.index = row_names

    if found_tuple["col"]:
        col_names = df.columns.tolist()
        col_names = [make_tuple(x) for x in col_names]
        df.columns = col_names

    # convert numbers to string if necessary
    #################################################
    if found_number["row"]:
        row_names = df.index.tolist()
        row_names = [str(x) for x in row_names]
        df.index = row_names

    if found_number["col"]:
        col_names = df.columns.tolist()
        col_names = [str(x) for x in col_names]
        df.columns = col_names

    return df



def add_enrichr_cats(df, inst_rc, run_enrichr, num_terms=10):
    from copy import deepcopy

    tmp_gene_list = deepcopy(df.index.tolist())

    gene_list = (
        [inst_tuple[0] for inst_tuple in tmp_gene_list]
        if isinstance(tmp_gene_list[0], tuple)
        else tmp_gene_list
    )

    orig_gene_list = deepcopy(gene_list)

    # set up for non-tuple case first
    if ": " in gene_list[0]:
        # strip titles
        gene_list = [inst_gene.split(": ")[1] for inst_gene in gene_list]

    # strip extra information (e.g. PTMs)
    gene_list = [inst_gene.split("_")[0] for inst_gene in gene_list]
    gene_list = [inst_gene.split(" ")[0] for inst_gene in gene_list]
    gene_list = [inst_gene.split("-")[0] for inst_gene in gene_list]

    user_list_id = post_request(gene_list)

    enr, response_list = get_request(run_enrichr, user_list_id, max_terms=20)

    # p-value, adjusted pvalue, z-score, combined score, genes
    # 1: Term
    # 2: P-value
    # 3: Z-score
    # 4: Combined Score
    # 5: Genes
    # 6: pval_bh

    # while generating categories store as list of lists, then convert to list of
    # tuples

    bar_info = []
    cat_list = [[inst_gene] for inst_gene in orig_gene_list]

    for inst_enr in response_list[:num_terms]:
        inst_term = inst_enr[1]
        inst_pval = inst_enr[2]
        inst_cs = inst_enr[4]
        inst_list = inst_enr[5]

        pval_string = f"<p> Pval {inst_pval}</p>"

        bar_info.append(inst_cs)

        for inst_info in cat_list:
            # strip titles
            gene_name = inst_info[0]

            if ": " in gene_name:
                gene_name = gene_name.split(": ")[1]

            # strip extra information (e.g. PTMs)
            gene_name = gene_name.split("_")[0]
            gene_name = gene_name.split(" ")[0]
            gene_name = gene_name.split("-")[0]

            result_str = (
                f"{inst_term}: True{pval_string}"
                if gene_name in inst_list
                else f"{inst_term}: False{pval_string}"
            )
            inst_info.append(result_str)

    cat_list = [tuple(x) for x in cat_list]

    df.index = cat_list

    return df, bar_info


def clust_from_response(response_list):
    from copy import deepcopy
    import math

    from clustergrammer import Network
    import pandas as pd
    import scipy

    # print('----------------------')
    # print('enrichr_clust_from_response')
    # print('----------------------')

    ini_enr = transfer_to_enr_dict(response_list)

    enr = []
    scores = {}
    score_types = ["combined_score", "pval", "zscore"]

    for score_type in score_types:
        scores[score_type] = pd.Series()

    for inst_enr in ini_enr:
        if inst_enr["combined_score"] > 0:
            # make series of enriched terms with scores
            for score_type in score_types:
                # collect the scores of the enriched terms
                if score_type == "combined_score":
                    scores[score_type][inst_enr["name"]] = inst_enr[score_type]
                elif score_type == "pval":
                    scores[score_type][inst_enr["name"]] = -math.log(inst_enr[score_type])
                elif score_type == "zscore":
                    scores[score_type][inst_enr["name"]] = -inst_enr[score_type]

            # keep enrichment values
            enr.append(inst_enr)

    # sort and normalize the scores
    for score_type in score_types:
        scores[score_type] = scores[score_type] / scores[score_type].max()
        scores[score_type].sort_values(ascending=False)

    number_of_enriched_terms = len(scores["combined_score"])

    enr_score_types = ["combined_score", "pval", "zscore"]

    if number_of_enriched_terms < 10:
        num_dict = {"ten": 10}
    elif number_of_enriched_terms < 20:
        num_dict = {"ten": 10, "twenty": 20}
    else:
        num_dict = {"ten": 10, "twenty": 20, "thirty": 30}

    # gather lists of top scores
    top_terms = {}
    for enr_type in enr_score_types:
        top_terms[enr_type] = {}
        for num_terms in list(num_dict.keys()):
            inst_num = num_dict[num_terms]
            top_terms[enr_type][num_terms] = scores[enr_type].index.tolist()[:inst_num]

    # gather the terms that should be kept - they are at the top of the score list
    keep_terms = []
    for inst_enr_score in top_terms:
        for tmp_num in list(num_dict.keys()):
            keep_terms.extend(top_terms[inst_enr_score][tmp_num])

    keep_terms = list(set(keep_terms))

    # keep enriched terms that are at the top 10 based on at least one score
    keep_enr = [inst_enr for inst_enr in enr if inst_enr["name"] in keep_terms]

    # fill in full matrix
    #######################

    # genes
    row_node_names = []
    # enriched terms
    col_node_names = []

    # gather information from the list of enriched terms
    for inst_enr in keep_enr:
        col_node_names.append(inst_enr["name"])
        row_node_names.extend(inst_enr["int_genes"])

    row_node_names = sorted(set(row_node_names))

    net = Network()
    net.dat["nodes"]["row"] = row_node_names
    net.dat["nodes"]["col"] = col_node_names
    net.dat["mat"] = scipy.zeros([len(row_node_names), len(col_node_names)])

    for inst_enr in keep_enr:
        inst_term = inst_enr["name"]
        col_index = col_node_names.index(inst_term)

        # use combined score for full matrix - will not be seen in viz
        tmp_score = scores["combined_score"][inst_term]
        net.dat["node_info"]["col"]["value"].append(tmp_score)

        for inst_gene in inst_enr["int_genes"]:
            row_index = row_node_names.index(inst_gene)

            # save association
            net.dat["mat"][row_index, col_index] = 1

    # cluster full matrix
    #############################
    # do not make multiple views
    views = [""]

    if len(net.dat["nodes"]["row"]) > 1:
        net.cluster(dist_type="jaccard", views=views, dendro=False)
    else:
        net.cluster(dist_type="jaccard", views=views, dendro=False, run_clustering=False)

    # get dataframe from full matrix
    df = net.dat_to_df()

    for score_type in score_types:
        for num_terms in num_dict:
            inst_df = deepcopy(df)
            inst_net = deepcopy(Network())

            inst_df = inst_df[top_terms[score_type][num_terms]]

            # load back into net
            inst_net.df_to_dat(inst_df)

            # make views
            if len(net.dat["nodes"]["row"]) > 1:
                inst_net.cluster(dist_type="jaccard", views=["N_row_sum"], dendro=False)
            else:
                inst_net.cluster(
                    dist_type="jaccard", views=["N_row_sum"], dendro=False, run_clustering=False
                )

            inst_views = inst_net.viz["views"]

            # add score_type to views
            for inst_view in inst_views:
                inst_view["N_col_sum"] = num_dict[num_terms]

                inst_view["enr_score_type"] = score_type

                # add values to col_nodes and order according to rank
                for inst_col in inst_view["nodes"]["col_nodes"]:
                    inst_col["rank"] = len(top_terms[score_type][num_terms]) - top_terms[
                        score_type
                    ][num_terms].index(inst_col["name"])

                    inst_name = inst_col["name"]
                    inst_col["value"] = scores[score_type][inst_name]

            # add views to main network
            net.viz["views"].extend(inst_views)

    return net


# make the get request to enrichr using the requests library
# this is done before making the get request with the lib name
def post_request(input_genes, meta=""):
    # get metadata
    import json

    import requests

    # stringify list
    input_genes = "\n".join(input_genes)

    # define post url
    post_url = "http://amp.pharm.mssm.edu/Enrichr/addList"

    # define parameters
    params = {"list": input_genes, "description": ""}

    # make request: post the gene list
    post_response = requests.post(post_url, files=params)

    # load json
    inst_dict = json.loads(post_response.text)
    # print(user_list_id)

    # return the user_list_id that is needed to reference the list later
    return str(inst_dict["userListId"])


# make the get request to enrichr using the requests library
# this is done after submitting post request with the input gene list
def get_request(lib, user_list_id, max_terms=50):
    import json

    import requests

    # convert user_list_id to string
    user_list_id = str(user_list_id)

    # define the get url
    get_url = "http://amp.pharm.mssm.edu/Enrichr/enrich"

    # get parameters
    params = {"backgroundType": lib, "userListId": user_list_id}

    # try get request until status code is 200
    inst_status_code = 400

    # wait until okay status code is returned
    num_try = 0

    # print(('\tEnrichr enrichment get req user_list_id: '+str(user_list_id)))

    while inst_status_code == 400 and num_try < 100:
        num_try += 1
        try:
            # make the get request to get the enrichr results

            try:
                get_response = requests.get(get_url, params=params)

                # get status_code
                inst_status_code = get_response.status_code

            except Exception:
                print("retry get request")

        except Exception:
            print("get requests failed")

    # load as dictionary
    resp_json = json.loads(get_response.text)

    # get the key
    only_key = next(iter(resp_json.keys()))

    # get response_list
    response_list = resp_json[only_key]

    # transfer the response_list to the enr_dict
    enr = transfer_to_enr_dict(response_list, max_terms)

    # return enrichment json and user_list_id
    return enr, response_list


# transfer the response_list to a list of dictionaries
def transfer_to_enr_dict(response_list, max_terms=50):
    # # reduce the number of enriched terms if necessary
    # if len(response_list) < num_terms:
    #   num_terms = len(response_list)

    # p-value, adjusted pvalue, z-score, combined score, genes
    # 1: Term
    # 2: P-value
    # 3: Z-score
    # 4: Combined Score
    # 5: Genes
    # 6: pval_bh

    num_enr_term = min(len(response_list), max_terms)

    # transfer response_list to enr structure
    # and only keep the top terms
    #
    # initialize enr
    enr = []
    for i in range(num_enr_term):
        # get list element
        inst_enr = response_list[i]

        # initialize dict
        inst_dict = {
            "name": inst_enr[1],  # transfer term
            "pval": inst_enr[2],  # transfer pval
            "zscore": inst_enr[3],  # transfer zscore
            "combined_score": inst_enr[4],  # transfer combined_score
            "int_genes": inst_enr[5],  # transfer int_genes
            "pval_bh": inst_enr[6],  # adjusted pval
        }

        # append dict
        enr.append(inst_dict)

    return enr



def main(net, inst_dm, which_sim, filter_sim, sim_mat_views=None):
    from copy import deepcopy

    from . import calc_clust
    from .__init__ import Network

    if sim_mat_views is None:
        sim_mat_views = ["N_row_sum"]

    sim_dict = {
        inst_rc: dm_to_sim(inst_dm[inst_rc], make_squareform=True, filter_sim=filter_sim)
        for inst_rc in which_sim
    }

    sim_net = {}

    for inst_rc in which_sim:
        sim_net[inst_rc] = deepcopy(Network())

        sim_net[inst_rc].dat["mat"] = sim_dict[inst_rc]

        sim_net[inst_rc].dat["nodes"]["row"] = net.dat["nodes"][inst_rc]
        sim_net[inst_rc].dat["nodes"]["col"] = net.dat["nodes"][inst_rc]

        sim_net[inst_rc].dat["node_info"]["row"] = net.dat["node_info"][inst_rc]
        sim_net[inst_rc].dat["node_info"]["col"] = net.dat["node_info"][inst_rc]

        calc_clust.cluster_row_and_col(sim_net[inst_rc])

        all_views = []
        _ = sim_net[inst_rc].dat_to_df()

        sim_net[inst_rc].viz["views"] = all_views

    return sim_net


def dm_to_sim(inst_dm, make_squareform=False, filter_sim=0):
    import numpy as np
    from scipy.spatial.distance import squareform

    if make_squareform:
        inst_dm = squareform(inst_dm)

    inst_sim_mat = 1 - inst_dm

    if filter_sim > 0:
        filter_sim = adjust_filter_sim(inst_sim_mat, filter_sim)
        inst_sim_mat[np.abs(inst_sim_mat) < filter_sim] = 0

    return inst_sim_mat


def adjust_filter_sim(inst_dm, filter_sim, keep_top=20000):
    import numpy as np
    import pandas as pd

    inst_df = pd.DataFrame(inst_dm)
    val_vect = np.abs(inst_df.values.flatten())

    val_vect = val_vect[val_vect > 0.01]

    if len(val_vect) > keep_top:
        inst_series = pd.Series(val_vect)
        inst_series.sort_values(ascending=False)

        sort_values = inst_series.values

        filter_sim = sort_values[keep_top]

    return filter_sim



# make_unique_labels


def main(net, df=None):
    """
    Run in load_data module (which runs when file is loaded or dataframe is loaded),
    check for duplicate row/col names, and add index to names if necessary
    """
    if df is None:
        df = net.export_df()

    # rows
    #############
    rows = df.index.tolist()
    if isinstance(rows[0], str):
        if len(rows) != len(list(set(rows))):
            print("warning: making row names unique")
            new_rows = add_index_list(rows)
            df.index = new_rows

    elif isinstance(rows[0], tuple):
        row_names = [inst_row[0] for inst_row in rows]

        if len(row_names) != len(list(set(row_names))):
            print("warning: making row names unique")
            row_names = add_index_list(row_names)

            # add back to tuple
            new_rows = []
            for inst_index in range(len(rows)):
                inst_row = rows[inst_index]
                new_row = list(inst_row)
                new_row[0] = row_names[inst_index]
                new_rows.append(tuple(new_row))

            df.index = new_rows

    # cols
    #############
    cols = df.columns.tolist()
    if isinstance(cols[0], str):
        # list column names
        if len(cols) != len(list(set(cols))):
            print("warning: making col names unique")
            new_cols = add_index_list(cols)
            df.columns = new_cols

    elif isinstance(cols[0], tuple):
        col_names = [inst_col[0] for inst_col in cols]

        if len(col_names) != len(list(set(col_names))):
            print("warning: making col names unique")
            col_names = add_index_list(col_names)

            # add back to tuple
            new_cols = []
            for inst_index in range(len(cols)):
                inst_col = cols[inst_index]
                new_col = list(inst_col)
                new_col[0] = col_names[inst_index]
                new_cols.append(tuple(new_col))

            df.columns = new_cols

    # return dataframe with unique names
    return df


def add_index_list(nodes):
    return [f"{nodes[i]}-{i + 1}" for i in range(len(nodes))]



import pandas as pd

from . import categories, make_unique_labels


def df_to_dat(net, df, define_cat_colors=False):
    """
    This is always run when data is loaded.
    """

    # check if df has unique values
    df = make_unique_labels.main(net, df)

    net.dat["mat"] = df.values
    net.dat["nodes"]["row"] = df.index.tolist()
    net.dat["nodes"]["col"] = df.columns.tolist()

    # print('checking ds status')
    # print('is_downsampled', net.is_downsampled)
    # print('meta_cat', net.meta_cat)
    # print(hasattr(net, 'meta_ds_col'))
    # print(hasattr(net, 'meta_ds_row'))

    # if net.meta_cat == False or net.is_downsampled:
    if not net.meta_cat:
        # tuple cats
        ##################################

        for axis in ["row", "col"]:
            inst_nodes = net.dat["nodes"][axis]

            if isinstance(inst_nodes[0], tuple):
                if axis == "row":
                    net.dat["node_info"][axis]["full_names"] = df.index.tolist()
                elif axis == "col":
                    net.dat["node_info"][axis]["full_names"] = df.columns.tolist()

                # get the number of categories from the length of the tuple
                # subtract 1 because the name is the first element of the tuple
                num_cats = len(inst_nodes[0]) - 1
                for inst_cat in range(num_cats):
                    cat_name = f"cat-{inst_cat}"
                    cat_index = inst_cat + 1
                    cat_values = [x[cat_index] for x in inst_nodes]
                    net.dat["node_info"][axis][cat_name] = cat_values

                # clean up nodes after parsing categories
                net.dat["nodes"][axis] = [x[0] for x in inst_nodes]

    else:
        # meta_cats
        ##########################

        for axis in ["row", "col"]:
            inst_nodes = net.dat["nodes"][axis]

            if axis == "row":
                net.dat["node_info"][axis]["full_names"] = df.index.tolist()
            elif axis == "col":
                net.dat["node_info"][axis]["full_names"] = df.columns.tolist()

            inst_cats = []
            if axis == "row":
                # inst_cats = net.meta_row.columns.tolist()
                if hasattr(net, "row_cats"):
                    inst_cats = net.row_cats
            else:
                # inst_cats = net.meta_col.columns.tolist()
                if hasattr(net, "col_cats"):
                    inst_cats = net.col_cats

            num_cats = len(inst_cats)

            # if axis == 'row':
            #   num_cats = len(net.row_cats)
            # elif axis == 'col':
            #   num_cats = len(net.col_cats)

            if num_cats > 0:
                for inst_cat in range(num_cats):
                    cat_name = f"cat-{inst_cat}"

                    cat_title = inst_cats[inst_cat]
                    if axis == "row":
                        if net.is_downsampled:
                            cat_values = (
                                net.meta_ds_row.loc[inst_nodes, cat_title]
                                .apply(lambda x, title=cat_title: f"{title}: {x}")
                                .values.tolist()
                                if hasattr(net, "meta_ds_row")
                                else net.meta_row.loc[inst_nodes, cat_title]
                                .apply(lambda x, title=cat_title: f"{title}: {x}")
                                .values.tolist()
                            )

                        # default with no downsampling
                        else:
                            cat_values = (
                                net.meta_row.loc[inst_nodes, cat_title]
                                .apply(lambda x, title=cat_title: f"{title}: {x}")
                                .values.tolist()
                            )
                    else:
                        # cat_values = net.meta_col.loc[inst_nodes, cat_title].apply(lambda x: cat_title + ': ' + x).values.tolist()
                        if net.is_downsampled:
                            cat_values = (
                                net.meta_ds_col.loc[inst_nodes, cat_title]
                                .apply(lambda x, title=cat_title: f"{title}: {x}")
                                .values.tolist()
                                if hasattr(net, "meta_ds_col")
                                else net.meta_col.loc[inst_nodes, cat_title]
                                .apply(lambda x, title=cat_title: f"{title}: {x}")
                                .values.tolist()
                            )

                        # default with no downsampling
                        else:
                            cat_values = (
                                net.meta_col.loc[inst_nodes, cat_title]
                                .apply(lambda x, title=cat_title: f"{title}: {x}")
                                .values.tolist()
                            )

                    net.dat["node_info"][axis][cat_name] = cat_values

    categories.dict_cat(net, define_cat_colors=define_cat_colors)


def dat_to_df(net):
    nodes = {
        axis: (
            net.dat["node_info"][axis]["full_names"]
            if "full_names" in net.dat["node_info"][axis]
            else net.dat["nodes"][axis]
        )
        for axis in ["row", "col"]
    }

    return pd.DataFrame(data=net.dat["mat"], columns=nodes["col"], index=nodes["row"])


def mat_to_numpy_arr(self):
    """convert list to numpy array - numpy arrays can not be saved as json"""
    import numpy as np

    self.dat["mat"] = np.asarray(self.dat["mat"])



"""Helper and utility functions."""

# Standard library imports
from collections.abc import Sequence
from typing import Any

# Third-party imports
import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import Point, base
from shapely.ops import transform


def _add_centroids_to_obsm(
    adata: Any,
    gdf: gpd.GeoDataFrame,
    key: str = "spatial",
) -> None:
    """
    Computes centroid x, y coordinates from a GeoDataFrame and stores them in adata.obsm.
    """
    if len(adata) != len(gdf):
        raise ValueError("Number of rows in adata and gdf must match.")

    centroids = gdf.geometry.centroid
    spatial_coords = np.vstack([centroids.x.values, centroids.y.values]).T
    adata.obsm[key] = spatial_coords


def _classify_polygons_contains_check(
    polygons: gpd.GeoSeries | Sequence[base.BaseGeometry],
    points: Sequence[Any],
) -> gpd.GeoSeries:
    """
    Classifies polygons as "real" or "fake" based on whether they contain any points inside.

    Parameters
    ----------
    polygons : GeoSeries of polygons (GeoPandas)
    points : Array-like of point coordinates (e.g., numpy array or list of tuples)

    Returns
    -------
    GeoSeries of curated polygons.
    """
    points_gdf = gpd.GeoDataFrame(geometry=[Point(p) for p in points])
    gdf_poly = gpd.GeoDataFrame(geometry=polygons)
    joined = gpd.sjoin(points_gdf, gdf_poly, predicate="within")
    real_polygons_indices = joined["index_right"].unique()
    curated_polygons = gdf_poly.iloc[real_polygons_indices]
    # Use .get() for a more concise and idiomatic way to handle the conditional return
    return curated_polygons.get("geometry", curated_polygons)


def _get_df_cell(adata: Any) -> pd.DataFrame:
    """
    Load cell-level cluster and spatial coordinates from an h5ad file as a DataFrame.
    """
    df_cell = pd.DataFrame(
        {
            "cluster": adata.obs["leiden"],
            "x": adata.obsm["spatial"][:, 0],
            "y": adata.obsm["spatial"][:, 1],
        }
    )
    df_cell["geometry"] = df_cell.apply(
        lambda row: [round(row["x"], 3), round(row["y"], 3)], axis=1
    )
    return df_cell


def _get_gdf_cell(adata: Any) -> gpd.GeoDataFrame:
    """
    Load cell-level cluster and spatial coordinates from an h5ad file as a GeoDataFrame.
    """
    return gpd.GeoDataFrame(
        {"cluster": adata.obs["leiden"]},
        geometry=gpd.points_from_xy(*adata.obsm["spatial"].T[:2]),
        crs="EPSG:4326",
    )


def _get_gdf_trx(data_dir: str) -> gpd.GeoDataFrame:
    """
    Load transcript data as a GeoDataFrame with spatial coordinates.
    """
    df_trx = pd.read_parquet(
        f"{data_dir}/transcripts.parquet",
        columns=["feature_name", "x_location", "y_location", "cell_id"],
        engine="pyarrow",
    )
    geometry = gpd.points_from_xy(df_trx["x_location"], df_trx["y_location"])
    return gpd.GeoDataFrame(df_trx[["feature_name", "cell_id"]], geometry=geometry, crs="EPSG:4326")


def _round_coordinates(
    geometry: base.BaseGeometry | None, precision: int = 2
) -> base.BaseGeometry | None:
    """
    Round the coordinates of a Shapely geometry to the specified precision.

    Parameters
    ----------
    geometry : Shapely geometry object (e.g., Polygon, MultiPolygon).
    precision : int
        Number of decimal places to round to.

    Returns
    -------
    Rounded Shapely geometry or None.
    """
    if geometry is None:
        return None

    def round_coords(
        x: float, y: float, z: float | None = None
    ) -> tuple[float, float] | tuple[float, float, float]:
        if z is not None:
            return (round(x, precision), round(y, precision), round(z, precision))
        return (round(x, precision), round(y, precision))

    return transform(round_coords, geometry)



from copy import deepcopy

import numpy as np
import pandas as pd


def main(net):
    """
    calculate pvalue of category closeness
    """
    # calculate the distance between the data points within the same category and
    # compare to null distribution
    for inst_rc in ["row", "col"]:
        inst_nodes = deepcopy(net.dat["nodes"][inst_rc])

        inst_index = deepcopy(net.dat["node_info"][inst_rc]["clust"])

        # reorder based on clustered order
        inst_nodes = [inst_nodes[i] for i in inst_index]

        # make distance matrix dataframe
        dm = dist_matrix_lattice(inst_nodes)

        node_infos = list(net.dat["node_info"][inst_rc].keys())

        all_cats = [inst_info for inst_info in node_infos if "dict_cat_" in inst_info]

        for cat_dict in all_cats:
            tmp_dict = net.dat["node_info"][inst_rc][cat_dict]

            pval_name = cat_dict.replace("dict_", "pval_")
            net.dat["node_info"][inst_rc][pval_name] = {}

            for cat_name in tmp_dict:
                subset = tmp_dict[cat_name]

                inst_median = calc_median_dist_subset(dm, subset)

                hist = calc_hist_distances(dm, subset, inst_nodes)

                pval = 0

                for i in range(len(hist["prob"])):
                    if i == 0:
                        pval = hist["prob"][i]
                    elif inst_median >= hist["bins"][i]:
                        pval = pval + hist["prob"][i]

                net.dat["node_info"][inst_rc][pval_name][cat_name] = pval


def dist_matrix_lattice(names):
    from scipy.spatial.distance import pdist, squareform

    lattice_size = len(names)
    mat = np.zeros([lattice_size, 1])
    mat[:, 0] = list(range(lattice_size))

    inst_dm = pdist(mat, metric="euclidean")

    inst_dm[inst_dm < 0] = float(0)

    inst_dm = squareform(inst_dm)

    return pd.DataFrame(data=inst_dm, columns=names, index=names)


def calc_median_dist_subset(dm, subset):
    return np.median(dm[subset].loc[subset].values)


def calc_hist_distances(dm, subset, inst_nodes):
    np.random.seed(100)

    num_null = 1000
    num_points = len(subset)

    median_dist = []
    for _ in range(num_null):
        tmp = np.random.choice(inst_nodes, num_points, replace=False)
        median_dist.append(np.median(dm[tmp].loc[tmp].values))

    median_dist = np.asarray(median_dist)
    s1 = pd.Series(median_dist)
    hist = np.histogram(s1, bins=30)

    return {
        "prob": hist[0] / float(num_null),
        "bins": hist[1],
    }



def check_categories(lines):
    """
    find out how many row and col categories are available
    when loading data from file
    """
    # count the number of row categories
    rcat_line = lines[0].split("\t")

    # calc the number of row names and categories
    num_rc = 0
    found_end = False

    # skip first tab
    for inst_string in rcat_line[1:]:
        if inst_string == "":
            if not found_end:
                num_rc = num_rc + 1
        else:
            found_end = True

    max_rcat = min(15, len(lines) - 1)

    num_cc = 0
    for i in range(max_rcat):
        ccat_line = lines[i + 1].split("\t")

        # make sure that line has length greater than one to prevent false cats from
        # trailing new lines at end of matrix
        if ccat_line[0] == "" and len(ccat_line) > 1:
            num_cc = num_cc + 1

    return {
        "row": num_rc + 1,
        "col": num_cc + 1,
    }


def dict_cat(net, define_cat_colors=False):
    """
    make a dictionary of node-category associations
    """
    net.persistent_cat_colors = True

    for inst_rc in ["row", "col"]:
        inst_keys = list(net.dat["node_info"][inst_rc].keys())
        all_cats = [x for x in inst_keys if "cat-" in x]

        for inst_name_cat in all_cats:
            dict_cat = {}
            tmp_cats = net.dat["node_info"][inst_rc][inst_name_cat]
            tmp_nodes = net.dat["nodes"][inst_rc]

            for i in range(len(tmp_cats)):
                inst_cat = tmp_cats[i]
                inst_node = tmp_nodes[i]

                if inst_cat not in dict_cat:
                    dict_cat[inst_cat] = []

                dict_cat[inst_cat].append(inst_node)

            tmp_name = f"dict_{inst_name_cat.replace('-', '_')}"
            net.dat["node_info"][inst_rc][tmp_name] = dict_cat

    # merge with old cat_colors by default
    cat_colors = net.viz["cat_colors"]

    if define_cat_colors:
        cat_number = 0

        for inst_rc in ["row", "col"]:
            inst_keys = list(net.dat["node_info"][inst_rc].keys())
            all_cats = [x for x in inst_keys if "cat-" in x]

            for cat_index in all_cats:
                if cat_index not in cat_colors[inst_rc]:
                    cat_colors[inst_rc][cat_index] = {}

                cat_names = sorted(set(net.dat["node_info"][inst_rc][cat_index]))

                # loop through each category name and assign a color
                for tmp_name in cat_names:
                    # using the same rules as the front-end to define cat_colors
                    inst_color = get_cat_color(cat_number + cat_names.index(tmp_name))

                    check_name = tmp_name

                    # check if category is string type and non-numeric
                    is_string_cat = not is_number(check_name)

                    if is_string_cat:
                        # check for default non-color
                        if ": " in check_name:
                            check_name = check_name.split(": ")[1]

                        # if check_name == 'False' or check_name == 'false':
                        if "False" in check_name or "false" in check_name:
                            inst_color = "#eee"

                        if "Not " in check_name:
                            inst_color = "#eee"

                    # do not overwrite old colors
                    if tmp_name not in cat_colors[inst_rc][cat_index] and is_string_cat:
                        cat_colors[inst_rc][cat_index][tmp_name] = inst_color
                        # print('overwrite: ' + tmp_name + ' -> ' + str(inst_color))

                    cat_number = cat_number + 1

        net.viz["cat_colors"] = cat_colors

        net.viz["global_cat_colors"] = {}

        global_cat_colors = {}
        for axis in net.viz["cat_colors"]:
            for cat_index in net.viz["cat_colors"][axis]:
                inst_dict = net.viz["cat_colors"][axis][cat_index]

                for inst_full_name in inst_dict:
                    inst_name = inst_full_name.split(": ")[1]
                    inst_color = inst_dict[inst_full_name]

                    global_cat_colors[inst_name] = inst_color

        net.viz["global_cat_colors"] = global_cat_colors


def calc_cat_clust_order(net, inst_rc):
    """
    cluster category subset of data
    """

    inst_keys = list(net.dat["node_info"][inst_rc].keys())

    if all_cats := [x for x in inst_keys if "cat-" in x]:
        for inst_name_cat in all_cats:
            tmp_name = f"dict_{inst_name_cat.replace('-', '_')}"
            dict_cat = net.dat["node_info"][inst_rc][tmp_name]

            unordered_cats = dict_cat.keys()

            ordered_cats = order_categories(unordered_cats)

            # this is the ordering of the columns based on their category, not
            # including their clustering ordering within category
            tmp_names_list = []
            for inst_cat in ordered_cats:
                inst_nodes = dict_cat[inst_cat]
                tmp_names_list.extend(inst_nodes)

            names_clust_list = tmp_names_list

            # calc category-cluster order
            final_order = []

            for i in range(len(net.dat["nodes"][inst_rc])):
                inst_node_name = net.dat["nodes"][inst_rc][i]
                inst_node_num = names_clust_list.index(inst_node_name)

                final_order.append(inst_node_num)

            inst_index_cat = f"{inst_name_cat.replace('-', '_')}_index"

            net.dat["node_info"][inst_rc][inst_index_cat] = final_order


def order_categories(unordered_cats):
    """
    If categories are strings, then simple ordering is fine.
    If categories are values then I'll need to order based on their values.
    The final ordering is given as the original categories (including titles) in a
    ordered list.
    """

    no_titles = remove_titles(unordered_cats)

    all_are_numbers = check_all_numbers(no_titles)

    return (
        order_cats_based_on_values(unordered_cats, no_titles)
        if all_are_numbers
        else sorted(unordered_cats)
    )


def order_cats_based_on_values(unordered_cats, values_list):
    import pandas as pd

    try:
        # convert values_list to values
        values_list = [float(i) for i in values_list]

        inst_series = pd.Series(data=values_list, index=unordered_cats)

        inst_series.sort_values(inplace=True)

        return inst_series.index.tolist()

    except Exception:
        # keep default ordering if error occurs
        print("error sorting cats based on values ")
        return unordered_cats


def check_all_numbers(no_titles):
    return all(is_number(tmp) for tmp in no_titles)


def remove_titles(cats):
    # check if all have titles
    ###########################
    all_have_titles = all(
        not is_number(inst_cat) and ": " in inst_cat for inst_cat in cats if not is_number(inst_cat)
    ) and any(not is_number(inst_cat) for inst_cat in cats)

    return [i.split(": ")[1] for i in cats] if all_have_titles else cats


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def get_cat_color(cat_num):
    all_colors = [
        "#393b79",
        "#aec7e8",
        "#ff7f0e",
        "#ffbb78",
        "#98df8a",
        "#bcbd22",
        "#404040",
        "#ff9896",
        "#c5b0d5",
        "#8c564b",
        "#1f77b4",
        "#5254a3",
        "#FFDB58",
        "#c49c94",
        "#e377c2",
        "#7f7f7f",
        "#2ca02c",
        "#9467bd",
        "#dbdb8d",
        "#17becf",
        "#637939",
        "#6b6ecf",
        "#9c9ede",
        "#d62728",
        "#8ca252",
        "#8c6d31",
        "#bd9e39",
        "#e7cb94",
        "#843c39",
        "#ad494a",
        "#d6616b",
        "#7b4173",
        "#a55194",
        "#ce6dbd",
        "#de9ed6",
    ]

    return all_colors[cat_num % len(all_colors)]


def add_cats(net, axis, cat_data):
    try:
        df = net.export_df()

        labels = df.index.tolist() if axis == "row" else df.columns.tolist()

        inst_title = cat_data.get("title", "New Category")

        all_cats = cat_data["cats"]

        # loop through all labels
        new_labels = []
        for inst_label in labels:
            if isinstance(inst_label, tuple):
                check_name = inst_label[0]
                found_tuple = True
            else:
                check_name = inst_label
                found_tuple = False

            if ": " in check_name:
                check_name = check_name.split(": ")[1]

            # default to False for found cat, overwrite if necessary
            found_cat = f"{inst_title}: False"

            # check all categories in cats
            for inst_cat in all_cats:
                inst_names = all_cats[inst_cat]

                if check_name in inst_names:
                    found_cat = f"{inst_title}: {inst_cat}"

            # add category to label
            new_label = (*inst_label, found_cat) if found_tuple else (inst_label, found_cat)

            new_labels.append(new_label)

        # add labels back to DataFrame
        if axis == "row":
            df.index = new_labels
        elif axis == "col":
            df.columns = new_labels

        net.load_df(df)

    except Exception:
        print("error adding new categories")



def df_filter_row_sum(df, threshold, take_abs=True):
    """filter rows in matrix at some threshold
    and remove columns that have a sum below this threshold"""

    from copy import deepcopy

    from .__init__ import Network

    _ = Network()

    df_copy = deepcopy(df.abs()) if take_abs else deepcopy(df)

    ini_rows = df_copy.index.values.tolist()
    df_copy = df_copy.transpose()
    tmp_sum = df_copy.sum(axis=0)
    tmp_sum = tmp_sum.abs()
    tmp_sum.sort_values(inplace=True, ascending=False)

    tmp_sum = tmp_sum[tmp_sum > threshold]
    keep_rows = sorted(tmp_sum.index.values.tolist())

    if len(keep_rows) < len(ini_rows):
        df = grab_df_subset(df, keep_rows=keep_rows)

    return df


def df_filter_col_sum(df, threshold, take_abs=True):
    """filter columns in matrix at some threshold
    and remove rows that have all zero values"""

    from copy import deepcopy

    from .__init__ import Network

    _ = Network()

    df_copy = deepcopy(df.abs()) if take_abs else deepcopy(df)

    df_copy = df_copy.transpose()
    df_copy = df_copy[df_copy.sum(axis=1) > threshold]
    df_copy = df_copy.transpose()
    df_copy = df_copy[df_copy.sum(axis=1) > 0]

    if take_abs:
        inst_rows = df_copy.index.tolist()
        inst_cols = df_copy.columns.tolist()
        df = grab_df_subset(df, inst_rows, inst_cols)

    else:
        df = df_copy

    return df


def grab_df_subset(df, keep_rows="all", keep_cols="all"):
    if keep_cols != "all":
        df = df[keep_cols]
    if keep_rows != "all":
        df = df.loc[keep_rows]
    return df


def get_sorted_rows(df, rank_type="sum"):
    from copy import deepcopy

    inst_df = deepcopy(df)
    inst_df = inst_df.transpose()

    tmp_sum = inst_df.sum(axis=0) if rank_type == "sum" else inst_df.var(axis=0)

    tmp_sum = tmp_sum.abs()
    tmp_sum.sort_values(inplace=True, ascending=False)
    return tmp_sum.index.values.tolist()


def filter_n_top(inst_rc, df, n_top, rank_type="sum"):
    if inst_rc == "col":
        df = df.transpose()

    rows_sorted = get_sorted_rows(df, rank_type)

    keep_rows = rows_sorted[:n_top]

    df = df.loc[keep_rows]

    if inst_rc == "col":
        df = df.transpose()

    return df


def filter_threshold(df, inst_rc, threshold, num_occur=1):
    """
    Filter a network's rows or cols based on num_occur values being above a
    threshold (in absolute_value)
    """
    from copy import deepcopy

    inst_df = deepcopy(df)

    if inst_rc == "col":
        inst_df = inst_df.transpose()

    inst_df = inst_df.abs()

    ini_rows = inst_df.index.values.tolist()

    inst_df[inst_df < threshold] = 0
    inst_df[inst_df >= threshold] = 1

    tmp_sum = inst_df.sum(axis=1)

    tmp_sum = tmp_sum[tmp_sum >= num_occur]

    keep_names = tmp_sum.index.values.tolist()

    if inst_rc == "row":
        if len(keep_names) < len(ini_rows):
            df = grab_df_subset(df, keep_rows=keep_names)

    elif inst_rc == "col":
        inst_df = inst_df.transpose()

        inst_rows = inst_df.index.values.tolist()
        inst_cols = keep_names

        df = grab_df_subset(df, inst_rows, inst_cols)

    return df


def filter_cat(net, axis, cat_index, cat_name):
    try:
        df = net.export_df()

        # DataFrame filtering will be run always be run on columns if the user
        # wants to filter rows, transpose the matrix before and after
        if axis == "row":
            df = df.transpose()

        all_names = df.columns.tolist()

        if found_names := [i for i in all_names if i[cat_index] == cat_name]:
            df = df[found_names]

            if axis == "row":
                df = df.transpose()
        else:
            print(f"no {axis}s were found with this category and filtering was not run")

        net.load_df(df)

    except Exception:
        print(
            "category filtering did not run\n check that your category filtering is set up correctly"
        )


def filter_names(net, axis, names):
    print("filter_names")
    print(names)

    try:
        df = net.export_df()

        # Dataframe filtering will always be run on the columns. If the user wants to filter rows, then it will transpose back and forth.

        if axis == "row":
            df = df.transpose()

        all_names = df.columns.tolist()

        found_names = []
        for inst_name in all_names:
            check_name = inst_name[0] if isinstance(inst_name, tuple) else inst_name

            if ": " in check_name:
                check_name = check_name.split(": ")[1]

            if check_name in names:
                found_names.append(inst_name)

        if found_names:
            df = df[found_names]

            if axis == "row":
                df = df.transpose()

            net.load_df(df)

        else:
            print(f"no {axis}s were found with these names")

    except Exception:
        print("error in filtering names")

    print(found_names)
